import logging

logger = logging.getLogger(__name__)


def check_argument(argument, name, types, columns=None):
    if not isinstance(argument, types):
        raise TypeError(f'Expected argument type {types} for variable {name}, got {type(argument)}.')

    # Check columns
    if columns is not None:
        if isinstance(columns, str):
            columns = [columns]
        missing = ', '.join([col for col in columns if col not in argument.columns])
        if any(missing):
            raise KeyError(f'Missing column(s) "{missing}"')

def warn_dataframe_not_empty(dataframe):
    if not dataframe.empty:
        logger.warning('The dataframe is not empty. The data are overwritten.')

def check_dictionary(dct, required, choice=[]):
    if required is not None:
        if isinstance(required, str):
            required = [required]
        for key in required:
            if not key in dct.keys():
                raise KeyError(f'Key "{key}" missing from dictionary.')

    if any(choice):
        for i, option in enumerate(choice):
            if not isinstance(option, (list, set)):
                choice[i] = [option]
        
        if not any([all([key in dct.keys() for key in option]) for option in choice]):
            options = [', '.join(option) for option in choice]
            raise KeyError('One of the following key sets should be present in the dictionary:\n - '.format('\n - '.join(options)))
            



import itertools
import logging
import os

import geopandas as gpd
import numpy as np
import pandas as pd
from tqdm.auto import tqdm
from scipy.spatial import KDTree
from shapely.geometry import LineString, Point, Polygon

from delft3dfmpy.converters import hydamo_to_dflowfm
from delft3dfmpy.core import checks, geometry
from delft3dfmpy.datamodels.common import ExtendedGeoDataFrame
from delft3dfmpy.datamodels.cstructures import meshgeom, meshgeomdim
from delft3dfmpy.io import dfmreader
logger = logging.getLogger(__name__)

class DFlowFMModel:
    """Main data structure for dflowfm model. Contains subclasses
    for network, structures, cross sections, observation points
    and external forcings.
    """

    def __init__(self):

        
        self.mdu_parameters = {}

        self.network = Network(self)

        self.structures = Structures(self)
        
        self.crosssections = CrossSections(self)

        self.observation_points = ObservationPoints(self)

        self.external_forcings = ExternalForcings(self)

    def export_network(self, output_dir, overwrite=False):
        """
        Expert network to shapefiles
        
        Three files are exported. One for the 2d mesh, one for the 1d mesh
        and one for the links between the 1d and 2d mesh.
    
        Parameters
        ----------
        output_dir : str
            Path to output directory
        overwrite : bool, optional
            Whether to overwrite existing files, by default False
        
        Raises
        ------
        FileExistsError
            If one of the shape file names already exists.
        """
        # Check if files already exist
        files = ['mesh1d.shp', 'mesh2d.shp', 'links1d2d.shp']
        paths = [os.path.join(output_dir, file) for file in files]
        if not overwrite:
            for path in paths:
                if os.path.exists(path):
                    raise FileExistsError(f'Path "{path}" already exists. Choose another output folder or specify overwrite=True.')

        # Links
        links = self.network.links1d2d.get_1d2dlinks(as_gdf=True)
        links.crs = {'init': 'epsg:28992'}
        links.to_file(paths[2])

        # Mesh2d
        mesh2d = gpd.GeoDataFrame(geometry=[Polygon(poly) for poly in self.network.mesh2d.get_faces()], crs='epsg:28992')
        # Add properties
        # Save
        mesh2d.to_file(paths[1])

        # Mesh1d
        mesh1d = gpd.GeoDataFrame(geometry=[LineString(line) for line in self.network.mesh1d.get_segments()], crs='epsg:28992')
        # Add properties
        mesh1d['node1'], mesh1d['node2'] = self.network.mesh1d.get_values('edge_nodes', as_array=True).T
        # Save
        mesh1d.to_file(paths[0])

class ExternalForcings:
    """
    Class for external forcings, which contains the boundary
    conditions and the initial conditions.
    """

    def __init__(self, dflowfmmodel):
        # Point to relevant attributes from parent
        self.dflowfmmodel = dflowfmmodel
        self.initial_waterlevel_polygons = gpd.GeoDataFrame(columns=['waterlevel', 'geometry'])
        self.initial_waterdepth_polygons = gpd.GeoDataFrame(columns=['waterdepth', 'geometry'])
        self.initial_waterlevel_xyz = []
        self.missing = None
        self.mdu_parameters = dflowfmmodel.mdu_parameters

        # GeoDataFrame for saving boundary conditions
        self.boundaries = {} #gpd.GeoDataFrame()
        
        # Dataframe for saving time series for structure
        self.structures = pd.DataFrame(columns=['id', 'type', 'parameter', 'time', 'value'])

        # Dictionary for saving laterals
        self.laterals = {}

        self.io = dfmreader.ExternalForcingsIO(self)

    def set_initial_waterlevel(self, level, polygon=None, name=None):
        """
        Method to set initial water level. A polygon can be given to
        limit the initial water level to a certain extent. 

        """
        # Get name is not given as input
        if name is None:
            name = 'wlevpoly{:04d}'.format(len(self.initial_waterlevel_polygons) + 1)

        # Add to geodataframe
        self.initial_waterlevel_polygons.loc[name] = {'waterlevel': level, 'geometry': polygon}

    def set_missing_waterlevel(self, missing):
        """
        Method to set the missing value for the water level.
        this overwrites the water level at missing value in the mdu file.

        Parameters
        ----------
        missing : float
            Water depth
        """
        self.mdu_parameters['WaterLevIni'] = missing
    
    def set_initial_waterdepth(self, depth, polygon=None, name=None):
        """
        Method to set the initial water depth in the 1d model. The water depth is
        set by determining the water level at the locations of the cross sections.
        
        Parameters
        ----------
        depth : float
            Water depth
        """
         # Get name is not given as input
        if name is None:
            name = 'wlevpoly{:04d}'.format(len(self.initial_waterlevel_polygons) + 1)
        # Add to geodataframe
        if polygon==None:
            
            new_df = pd.DataFrame({'waterdepth': depth, 'geometry': polygon}, index=[name])
            
            self.initial_waterdepth_polygons =  new_df
        else:
            self.initial_waterdepth_polygons.loc[name] = {'waterdepth': depth, 'geometry': polygon}
        
    def add_rainfall_2D(self, fName, bctype='rainfall'):
        """
        Parameters
        ----------
        fName : str
            Location of netcdf file containing rainfall rasters
        bctype : str
            Type of boundary condition. Currently only rainfall is supported
        """
        
        assert bctype in ['rainfall']
        
        # Add boundary condition
        self.boundaries['rainfall_2D'] = {
            'file_name': fName,
            'bctype': bctype+'bnd',
        }
        

    def add_boundary_condition(self, name, pt, bctype, series, branchid=None):
        """
        Add boundary conditions to model:
        - The boundary condition can be discharge or waterlevel
        - Is specified by a geographical location (pt) and a branchid
        - If no branchid is given, the nearest is searched
        - The boundary condition is added to the end of the given or nearest branch.
        - To specify a time dependendend boundary: a timeseries with values should be given
        - To specify a constant boundary a float should be given
        
        Parameters
        ----------
        name : str
            ID of the boundary condition
        pt : tuple or shapely.geometry.Point
            Location of the boundary condition
        bctype : str
            Type of boundary condition. Currently only discharge and waterlevel are supported
        series : pd.Series or float
            If a float, a constant in time boundary condition is used. If a pandas series,
            the values per time step are used. Index should be in datetime format
        branchid : str, optional
            ID of the branch. If None, the branch nearest to the given location (pt) is
            searched, by default None
        """

        assert bctype in ['discharge', 'waterlevel']

        unit = 'm3/s' if bctype=='discharge' else 'm'
        if name in self.boundaries.keys():
            raise KeyError(f'A boundary condition with name "{name}" is already present.')
        
        if isinstance(pt, tuple):
            pt = Point(*pt)

        # Find the nearest node        
        nodes1d = np.asarray([n.split('_') for n in self.dflowfmmodel.network.mesh1d.description1d['network_node_ids']], dtype='float')
        get_nearest = KDTree(nodes1d)
        distance, idx_nearest = get_nearest.query(pt)
        nodeid = f'{nodes1d[idx_nearest][0]:g}_{nodes1d[idx_nearest][1]:g}'
        
        # # If branch is not given explicitly, find the nearest
        # if branchid is None:
        #     branchid = self.dflowfmmodel.network.branches.distance(pt).idxmin()

        # # Find nearest branch for geometry
        # branch = self.dflowfmmodel.network.schematised.at[branchid, 'geometry']
        # extended_line = geometry.extend_linestring(line=branch, near_pt=pt, length=1.0)

        # # Create intersection line for boundary condition
        # bcline = LineString(geometry.orthogonal_line(line=extended_line, offset=0.1, width=0.1))

        # Convert time to minutes
        if isinstance(series, pd.Series):
            times = ((series.index - series.index[0]).total_seconds() / 60.).tolist()
            values = series.values.tolist()
        else:
            times = None
            values = series

        # Add boundary condition
        self.boundaries[name] = {
            'code': name,
            'bctype': bctype+'bnd',
            'value': values,
            'time': times,            
            'filetype': 9,
            'operand': 'O',
            'method': 3,
            'unit' : unit,            
            'nodeid': nodeid
        }
        
        # Check if a 1d2d link should be removed
        self.dflowfmmodel.network.links1d2d.check_boundary_link(self.boundaries[name])
    
    def add_rain_series(self, name, values, times):
        """
        Adds a rain series a boundary condition.
        Specify name, values, and times
        
        Parameters
        ----------
        name : str
            ID of the condition
        values : list of floats
            Values of the rain intensity
        times : list of datetime
            Times for the values
        """
        # Add boundary condition
        self.boundaries[name] = {
            'code' : name,
            'bctype' : 'rainfall',
            'filetype' : 1,
            'method' : 1,
            'operand' : 'O',
            'value': values,
            'time': times,
            'geometry': None,
            'branchid': None
        }

    def set_structure_series(self, structure_id, structure_type, parameter, times, values):
        # Get filename
        filename = f"{structure_type}_{structure_id}.tim"

        if structure_type == 'weir':
            weirnames = list(self.dflowfmmodel.structures.weirs.keys())
            if structure_id not in weirnames:
                raise IndexError(f'"{structure_id}" not in index: "{",".join(weirnames)}"')
            self.dflowfmmodel.structures.weirs[structure_id][parameter] = filename
        elif structure_type == 'orifice':
            orificenames = list(self.dflowfmmodel.structures.orifices.keys())
            if structure_id not in orificenames:
                raise IndexError(f'"{structure_id}" not in index: "{",".join(orificenames)}"')
            self.dflowfmmodel.structures.orifices[structure_id][parameter] = filename        
        else:
            raise NotImplementedError('Only implemented for weirs and orifices.')
    # Add in structure dataframe
     
        # Add boundary condition
        self.structures.loc[structure_id] = {
            'id' : structure_id,
            'type' : structure_type,
            'parameter' : parameter,
            'time': times,
            'value': values
        }

class CrossSections:
    """
    Cross section class.
    Contains dictionaries for the cross section locations and 
    cross section definitions. Also has an subclass for input
    and output.
    """
    def __init__(self, dflowfmmodel):
        """
        Constructor
        
        Parameters
        ----------
        DFlowFMMOdel
            Model schematisation of which the cross sections are part.
        """
        self.io = dfmreader.CrossSectionsIO(self)
        
        self.crosssection_loc = {}
        self.crosssection_def = {}

        self.dflowfmmodel = dflowfmmodel

        self.default_definition = None
        self.default_definition_shift = 0.0

        self.get_roughnessname = self.dflowfmmodel.network.get_roughness_description        
    
    def set_default_definition(self, definition, shift=0.0):
        """
        Add default profile
        """
        if definition not in self.crosssection_def.keys():
            raise KeyError(f'Cross section definition "{definition}" not found."')

        self.default_definition = definition
        self.default_definition_shift = shift

    def add_yz_definition(self, yz=None, thalweg=None, roughnesstype=None, roughnessvalue=None, name=None):
        """
        Add xyz crosssection

        Parameters
        ----------
        code : str
            Id of cross section
        branch : str
            Name of branch
        offset : float
            Position of cross section along branch. If not given, the position is determined
            from the branches in the network. These should thus be given in this case.
        crds : np.array
            Nx2 array with y, z coordinates
        """

        # get coordinates
        length, z = yz.T
        if name is None:
            name = f'yz_{len(crosssection_def):08d}'
        
        # Get roughnessname
        roughnessname = self.get_roughnessname(roughnesstype, roughnessvalue)

        # Add to dictionary
        self.crosssection_def[name] = {
            'id' : name,
            'type': 'yz',
            'thalweg': np.round(thalweg,decimals=3),
            'yzCount': len(z),
            'yCoordinates': list_to_str(length),
            'zCoordinates': list_to_str(z),
            'sectionCount': 1,
            'frictionIds': roughnessname,
            'frictionPositions': list_to_str([length[0], length[-1]])
        }

        return name

    def add_circle_definition(self, diameter, roughnesstype, roughnessvalue, name=None):
        """
        Add circle cross section. The cross section name is derived from the shape and roughness,
        so similar cross sections will result in a single definition.
        """        
        # Get name if not given
        if name is None:
            name = f'circ_d{diameter:.3f}'
        
        # Get roughnessname
        roughnessname = self.get_roughnessname(roughnesstype, roughnessvalue)

        # Add to dictionary
        self.crosssection_def[name] = {
            'id' : name,
            'type': 'circle',
            'thalweg': 0.0,
            'diameter': diameter,
            'frictionId': roughnessname
        }

        return name        

    def add_rectangle_definition(self, height, width, closed, roughnesstype, roughnessvalue, name=None):
        """
        Add rectangle cross section. The cross section name is derived from the shape and roughness,
        so similar cross sections will result in a single definition.
        """        
        # Get name if not given
        if name is None:
            name = f'rect_h{height:.3f}_w{width:.3f}'

        # Get roughnessname
        roughnessname = self.get_roughnessname(roughnesstype, roughnessvalue)

        # Add to dictionary
        self.crosssection_def[name] = {
            'id' : name,
            'type': 'rectangle',
            'thalweg': 0.0,
            'height': height,
            'width': width,
            'closed': int(closed),
            'frictionid': roughnessname
        }

        return name

    def add_trapezium_definition(self, slope, maximumflowwidth, bottomwidth, closed, roughnesstype, roughnessvalue, name=None):
        """
        Add rectangle cross section. The cross section name is derived from the shape and roughness,
        so similar cross sections will result in a single definition.
        """        
        # Get name if not given
        if name is None:
            name = f'trapz_s{slope:.1f}_bw{bottomwidth:.1f}_bw{maximumflowwidth:.1f}'
        
        # Get roughnessname
        roughnessname = self.get_roughnessname(roughnesstype, roughnessvalue)

        if not closed:
            levels = '0 100'
            flowwidths = f'{bottomwidth:.2f} {bottomwidth + 2 * slope * 100:.2f}'
        else:
            levels = f'0 {(maximumflowwidth - bottomwidth) / (2 * slope):.2f}'
            flowwidths = f'{bottomwidth:.2f} {maximumflowwidth:.2f}'

        # Add to dictionary
        self.crosssection_def[name] = {
            'id' : name,
            'type': 'zw',
            'thalweg': 0.0,
            'numLevels': 2,
            'levels': levels,
            'flowWidths': flowwidths,
            'totalWidths': flowwidths,
            'frictionId': roughnessname
        }

        return name

    def add_crosssection_location(self, branchid, chainage, definition, minz=np.nan, shift=0.0):

        descr = f'{branchid}_{chainage:.1f}'
        # Add cross section location
        self.crosssection_loc[descr] = {
            'id': descr,
            'branchid': branchid,
            'chainage': chainage,
            'shift': shift,
            'definitionId': definition,
        }

    def get_branches_without_crosssection(self):
        # First find all branches that match a cross section
        branch_ids = {dct['branchid'] for _, dct in self.crosssection_loc.items()}
        # Select the branch-ids that do nog have a matching cross section
        branches = self.dflowfmmodel.network.branches
        no_crosssection = branches.index[~np.isin(branches.index, list(branch_ids))]

        return no_crosssection.tolist()

    def get_structures_without_crosssection(self):
        struc_ids =  [dct['id'] for _, dct in self.crosssection_def.items()]
        bridge_ids = [dct['csDefId'] for _, dct in self.dflowfmmodel.structures.bridges.items()] 
        no_cross_bridge = np.asarray(bridge_ids)[~np.isin(bridge_ids , struc_ids)].tolist() 
        no_crosssection = no_cross_bridge        
        #uweir_ids = [dct['numlevels'] for _, dct in self.dflowfmmodel.structures.uweirs.items()]                
        #no_cross_uweir = uweirs.index[~np.isin(uweirs.index, list(uweir_ids))]        
        #no_crosssection = no_crosssection.append(no_cross_uweir)
        return no_crosssection

    def get_bottom_levels(self):
        """Method to determine bottom levels from cross sections"""

        # Initialize lists
        data = []
        geometry = []
        
        for key, css in self.crosssection_loc.items():
            # Get location
            geometry.append(self.dflowfmmodel.network.schematised.at[css['branchid'], 'geometry'].interpolate(css['chainage']))
            shift = css['shift']

            # Get depth from definition if yz and shift
            definition = self.crosssection_def[css['definitionId']]
            minz = shift
            if definition['type'] == 'yz':
                minz += min(float(z) for z in definition['zCoordinates'].split())
            
            data.append([css['branchid'], css['chainage'], minz])

        # Add to geodataframe
        gdf = gpd.GeoDataFrame(
            data=data,
            columns=['branchid', 'chainage', 'minz'],
            geometry=geometry
        )

        return gdf



class Links1d2d:

    def __init__(self, network):
        self.mesh1d = network.mesh1d
        self.mesh2d = network.mesh2d
        self.network = network

        # List for 1d 2d links
        self.nodes1d = []
        self.faces2d = []

    def generate_1d_to_2d(self, max_distance=np.inf, branchid=None):
        """
        Generate 1d2d links from 1d nodes. Each 1d node is connected to
        the nearest 2d cell. A maximum distance can be specified to remove links
        that are too long. Also the branchid can be specified, if you only want
        to generate links from certain 1d branches.
        
        Parameters
        ----------
        max_distance : int, float
            The maximum length of a link. All longer links are removed.
        branchid : str or list
            ID's of branches for which the connection from 1d to 2d is made.
        """
        logger.info(f'Generating links from 1d to 2d based on distance.')
        
        # Create KDTree for faces
        faces2d = np.c_[self.mesh2d.get_values('facex'), self.mesh2d.get_values('facey')]
        get_nearest = KDTree(faces2d)

        # Get network geometry
        nodes1d = self.mesh1d.get_nodes()
        idx = self.mesh1d.get_nodes_for_branch(branchid)

        # Get nearest 2d nodes
        distance, idx_nearest = get_nearest.query(nodes1d[idx])
        close = (distance < max_distance)
        
        # Add link data
        nodes1didx = np.arange(len(nodes1d))[idx]
        self.nodes1d.extend(nodes1didx[close] + 1)
        self.faces2d.extend(idx_nearest[close] + 1)

        # Remove conflicting 1d2d links
        for bc in self.network.dflowfmmodel.external_forcings.boundaries.values():
            if bc['geometry'] is None:
                continue
            self.check_boundary_link(bc)

    def generate_2d_to_1d(self, max_distance=np.inf, intersecting=True, branchid=None, shift_to_centroid=True):
        """
        Generate 1d2d links from 2d cells. A maximum distance can be specified
        to remove links that are too long. Also a branchid can be specified to only
        generate links to certain branches.
        
        In case of a 1D and 2D grid that is on top of each other the user might want
        to generate links only for intersecting cells, where in case of non-overlapping meshes
        the use might want to use the shortest distance. This behaviour can be specified with
        the option intersecting:
        1. intersecting = True: each 2d cell crossing a 1d branch segment is connected to
            the nearest 1d cell.
        2. intersecting = False: each 2d cell is connected to the nearest 1d cell,
            If the link crosses another cell it is removed.
        In case of option 2. setting a max distance will speed up the the process a bit.

        Parameters
        ----------
        max_distance : int, float
            Maximum allowed length for a link
        intersecting : bool
            Make connections for intersecting 1d and 2d cells or based on
            nearest neighbours
        branchid : str or list of str
            Generate only to specified 1d branches
        """
        logger.info(f'Generating links from 2d to 1d based on {"intersection" if intersecting else "distance"}.')

        # Collect polygons for cells
        centers2d = self.mesh2d.get_faces(geometry='center')
        idx = np.arange(len(centers2d), dtype='int')
        # Create KDTree for 1d cells
        nodes1d = self.mesh1d.get_nodes()
        nodes1didx = self.mesh1d.get_nodes_for_branch(branchid)
        get_nearest = KDTree(nodes1d[nodes1didx])
        
        # Make a pre-selection
        if max_distance < np.inf:
            # Determine distance from 2d to nearest 1d
            distance, _ = get_nearest.query(centers2d)
            idx = idx[distance < max_distance]
        
        # Create GeoDataFrame
        logger.info(f'Creating GeoDataFrame of ({len(idx)}) 2D cells.')
        cells = gpd.GeoDataFrame(
            data=centers2d[idx],
            columns=['x', 'y'],
            index=idx + 1,
            geometry=[Polygon(cell) for i, cell in enumerate(self.mesh2d.get_faces()) if i in idx]
        )
        
        # Find intersecting cells with branches
        logger.info('Determine intersecting or nearest branches.')
        if branchid is None:
            branches = self.network.branches
        elif isinstance(branchid, str):
            branches = self.network.branches.loc[[branchid]]
        else:
            branches = self.network.branches.loc[branchid]
        
        if intersecting:
            geometry.find_nearest_branch(branches, cells, method='intersecting')
        else:
            geometry.find_nearest_branch(branches, cells, method='overal', maxdist=max_distance)

        # Drop the cells without intersection
        cells.dropna(subset=['branch_offset'], inplace=True)
        faces2d = np.c_[cells.x, cells.y]
        
        # Get nearest 1d nodes
        distance, idx_nearest = get_nearest.query(faces2d)
        close = (distance < max_distance)
        
        # Add link data
        nodes1didx = np.where(nodes1didx)[0][idx_nearest]
        self.nodes1d.extend(nodes1didx[close] + 1)
        self.faces2d.extend(cells.index.values[close])

        if not intersecting:
            logger.info('Remove links that cross another 2D cell.')
            # Make sure only the nearest cells are accounted by removing all links that also cross another cell
            links = self.get_1d2dlinks(as_gdf=True)
            todrop = []

            # Remove links that intersect multiple cells
            cellbounds = cells.bounds.values.T
            for link in tqdm(links.itertuples(), total=len(links), desc='Removing links crossing mult. cells'):
                selectie = cells.loc[geometry.possibly_intersecting(cellbounds, link.geometry)].copy()
                if selectie.intersects(link.geometry).sum() > 1:
                    todrop.append(link.Index)
            links.drop(todrop, inplace=True)

            # Re-assign
            del self.nodes1d[:]
            del self.faces2d[:]

            self.nodes1d.extend(links['node1did'].values.tolist())
            self.faces2d.extend(links['face2did'].values.tolist())

        # Shift centers of 2d faces to centroid if they are part of a 1d-2d link
        if shift_to_centroid:
            # Get current centers
            cx, cy = self.mesh2d.get_faces(geometry='center').T
            # Calculate centroids for cells with link
            idx = np.array(self.faces2d) - 1
            centroids = np.vstack([cell.mean(axis=0) for cell in np.array(self.mesh2d.get_faces())[idx]]).T
            cx[idx] = centroids[0]
            cy[idx] = centroids[1]
            # Set values back to geometry
            self.mesh2d.set_values('facex', cx)
            self.mesh2d.set_values('facey', cy)

        # Remove conflicting 1d2d links
        for bc in self.network.dflowfmmodel.external_forcings.boundaries.values():
            if bc['geometry'] is None:
                continue
            self.check_boundary_link(bc)

    def check_boundary_link(self, bc):
        """
        Since a boundary conditions is not picked up when there is a bifurcation
        in the first branch segment, potential 1d2d links should be removed.

        This function should be called whenever a boundary conditions is added,
        or the 1d2d links are generated.
        """

        # Can only be done after links have been generated
        if not self.nodes1d or not self.faces2d:
            return None

        # Find the nearest node with the KDTree
        nodes1d = self.mesh1d.get_nodes()
        get_nearest = KDTree(nodes1d)
        distance, idx_nearest = get_nearest.query([float(pt) for pt in bc['nodeid'].split('_')])
        node_id = idx_nearest + 1

        # Check 1. Determine if the nearest node itself is not a bifurcation
        edge_nodes = self.mesh1d.get_values('edge_nodes', as_array=True)
        counts = {u: c for u, c in zip(*np.unique(edge_nodes, return_counts=True))}
        if counts[node_id] > 1:
            logger.warning(f'The boundary condition at {node_id} is not a branch end. Check if it is picked up by dflowfm.')

        # Check 2. Check if any 1d2d links are connected to the node or next node. If so, remove.
        # Find the node(s) connect to 'node_id'
        to_remove = np.unique(edge_nodes[(edge_nodes == node_id).any(axis=1)])
        for item in to_remove:
            while item in self.nodes1d:
                loc = self.nodes1d.index(item)
                self.nodes1d.pop(loc)
                self.faces2d.pop(loc)
                nx, ny = nodes1d[item-1]
                #bcx, bcy = bc['geometry'].centroid.coords[0]
                logger.info(f'Removed link(s) from 1d node: ({nx:.2f}, {ny:.2f}) because it is too close to boundary condition at node {node_id:.0f}.')
            
    def get_1d2dlinks(self, as_gdf=False):
        """
        Method to get 1d2d links as array with coordinates or geodataframe.

        Parameters
        ----------
        as_gdf : bool
            Whether to export as geodataframe (True) or numpy array (False)
        """

        if not any(self.nodes1d):
            return None

        # Get 1d nodes and 2d faces
        nodes1d = self.mesh1d.get_nodes()
        faces2d = self.mesh2d.get_faces(geometry='center')

        # Get links
        links = np.dstack([nodes1d[np.array(self.nodes1d) - 1], faces2d[np.array(self.faces2d) - 1]])

        if not as_gdf:
            return np.array([line.T for line in links])
        else:
            return gpd.GeoDataFrame(
                data=np.c_[self.nodes1d, self.faces2d],
                columns=['node1did', 'face2did'],
                geometry=[LineString(line.T) for line in links]
            )

    def remove_1d2d_from_numlimdt(self, file, threshold, node='2d'):
        """
        Remove 1d2d links based on numlimdt file
        """
        if node == '1d':
            links = self.get_1d2dlinks(as_gdf=True)

        with open(file) as f:
            for line in f.readlines():
                x, y, n = line.split()
                if int(n) >= threshold:
                    if node == '2d':
                        self.remove_1d2d_link(float(x), float(y), mesh=node, max_distance=2.0)
                    else:
                        # Find the 1d node connected to the link
                        idx = links.distance(Point(float(x), float(y))).idxmin()
                        x, y = links.at[idx, 'geometry'].coords[0]
                        self.remove_1d2d_link(x, y, mesh=node, max_distance=2.0)

    def remove_1d2d_link(self, x, y, mesh, max_distance):
        """
        Remove 1d 2d link based on x y coordinate.
        Mesh can specified, 1d or 2d.
        """
        if mesh == '1d':
            pts = self.mesh1d.get_nodes()
            ilink = 0
        elif mesh == '2d':
            pts = np.c_[self.mesh2d.get_faces(geometry='center')]
            ilink = 1
        else:
            raise ValueError()

        # Find nearest link
        dists = np.hypot(pts[:, 0] - x, pts[:, 1] - y)
        if dists.min() > max_distance:
            return None
        imin = np.argmin(dists)

        # Determine what rows to remove (if any)
        linkdim = self.nodes1d if mesh == '1d' else self.faces2d
        to_remove = [link for link in (linkdim) if link == (imin + 1)]
        for item in to_remove:
            while item in linkdim:
                loc = linkdim.index(item)
                self.nodes1d.pop(loc)
                self.faces2d.pop(loc)

class Network:

    def __init__(self, dflowfmmodel):
        # Link dflowmodel
        self.dflowfmmodel = dflowfmmodel

        # Mesh 1d offsets
        self.offsets = {}

        # Branches and schematised branches
        self.branches = ExtendedGeoDataFrame(geotype=LineString, required_columns=['code', 'geometry'])
        self.schematised = ExtendedGeoDataFrame(geotype=LineString, required_columns=['geometry'])

        # Create mesh for the 1d network
        self.mesh1d = meshgeom(meshgeomdim())
        self.mesh1d.meshgeomdim.dim = 1

        self.mesh1d.edge_branchoffsets = []
        self.mesh1d.edge_branchidx = []
        self.mesh1d.edge_x = []
        self.mesh1d.edge_y = []
                
        # Create mesh for the 1d network
        self.mesh2d = meshgeom(meshgeomdim())
        self.mesh2d.meshgeomdim.dim = 2

        # Create 1d2dlinks
        self.links1d2d = Links1d2d(self)

        # Dictionary for roughness definitions
        self.roughness_definitions = {}

        # Link mdu parameters
        self.mdu_parameters = dflowfmmodel.mdu_parameters
        
        
    def set_branches(self, branches):
        """
        Set branches from geodataframe
        """
        # Check input
        checks.check_argument(branches, 'branches', (ExtendedGeoDataFrame, gpd.GeoDataFrame))
        # Add data to branches
        self.branches.set_data(branches[self.branches.required_columns])
        # Copy branches to schematised
        self.schematised.set_data(self.branches[self.schematised.required_columns])

    def snap_branch_ends(self, offset):
        """
        Method to snap branch ends to other branch ends within a given offset.

        Parameters
        offset : float
            Maximum distance between end points. If the distance is larger, they are not snapped.
        """
        # Collect endpoints
        endpoints = []
        for branch in self.branches.itertuples():
            endpoints.append((branch.geometry.coords[0], branch.Index, 0))
            endpoints.append((branch.geometry.coords[-1], branch.Index, -1))

        # Create KDTree of endpoints
        snapped = 0

        # For every endpoint determine the distance to the nearest other endpoint
        for i, (endpoint, branchid, side) in enumerate(endpoints):
            # Find distance and nearest coordinate of other points
            other_pts = [pt[0] for j, pt in enumerate(endpoints) if j != i]
            mindist, minidx = KDTree(other_pts).query(endpoint)

            # Change endpoint if dist is not 0.0 but smaller than offset.
            if mindist != 0.0 and mindist <= offset:
               
                # Change coordinates of branch
                crds = self.branches.at[branchid, 'geometry'].coords[:]
                crds[side] = other_pts[minidx]
                self.branches.at[branchid, 'geometry'] = LineString(crds)
                snapped += 1

                endpoints[i] = (other_pts[minidx], branchid, side)

        print(f'Snapped {snapped} points.')

    def set_branch_order(self, branchids, idx=None):
        """
        Group branch ids so that the cross sections are
        interpolated along the branch.
        
        Parameters
        ----------
        branchids : list
            List of branches to group
        """
        # Get the ids (integers) of the branch names given by the user
        branchidx = np.isin(self.mesh1d.description1d['network_branch_ids'], branchids)
        # Get current order
        branchorder = self.mesh1d.get_values('nbranchorder', as_array=True)
        # Update
        if idx is None:
            branchorder[branchidx] = branchorder.max() + 1
        else:
            if not isinstance(idx, int):
                raise TypeError('Expected integer.')
            branchorder[branchidx] = idx
        # Save
        self.mesh1d.set_values('nbranchorder', branchorder)
        
    # generate network and 1d mesh
    def generate_1dnetwork(self, one_d_mesh_distance=40.0, seperate_structures=True, max_dist_to_struc=None):
        """
        Parameters
        ----------
        one_d_mesh_distance : float

        single_edge_nodes : boolean
        """

        if self.branches.empty:
            raise ValueError('Branches should be added before 1d network can be generated.')

        checks.check_argument(one_d_mesh_distance, 'one_d_mesh_distance', (float, int))

        # Temporary dictionary to store the id number of the nodes and branches
        nodes = []
        edge_nodes_dict = {}

        # Check if any structures present (if not, structures will be None)
        structures = self.dflowfmmodel.structures.as_dataframe(weirs=True, bridges=True, culverts=True, pumps=True, uweirs=True, orifices=True, compounds=True)

        # If offsets are not predefined, generate them base on one_d_mesh_distance
        if not self.offsets:
            self.generate_offsets(one_d_mesh_distance, structures=structures, max_dist_to_struc=max_dist_to_struc)

        # Add the network data to the 1d mesh structure
        sorted_branches = self.branches.iloc[self.branches.length.argsort().values]

        # Add network branch data
        dimensions = self.mesh1d.meshgeomdim
        dimensions.nbranches = len(sorted_branches)
        self.mesh1d.set_values('nbranchorder', (np.ones(dimensions.nbranches, dtype=int) * -1).tolist())
        self.mesh1d.set_values('nbranchlengths', sorted_branches.geometry.length + 1e-12)
        self.mesh1d.description1d['network_branch_ids'] = sorted_branches.index.astype(str).tolist()
        self.mesh1d.description1d['network_branch_long_names'] = sorted_branches.index.astype(str).tolist()
        
        # Add network branch geometry
        coords = [line.coords[:] for line in sorted_branches.geometry]
        geomx, geomy = list(zip(*list(itertools.chain(*coords))))[:2]
        dimensions.ngeometry = len(geomx)
        self.mesh1d.set_values('nbranchgeometrynodes', [len(lst) for lst in coords])
        self.mesh1d.set_values('ngeopointx', geomx)
        self.mesh1d.set_values('ngeopointy', geomy)

        branch_names = sorted_branches.index.astype(str).tolist()
        branch_longnames = ['long_' + s for s in branch_names]

        network_edge_nodes = []
        mesh1d_edge_nodes = []
        mesh1d_node_branchidx = []
        mesh1d_node_branchoffset = []        
        mesh1d_edge_branchidx = []
        mesh1d_edge_branchoffset = []
        mesh1d_edge_x = []
        mesh1d_edge_y = []       
        mesh1d_node_names = []
                   
        # For each branch
        for i_branch, branch in enumerate(sorted_branches.itertuples()):

            # Get branch coordinates
            points = branch.geometry.coords[:]

            # Network edge node administration
            # -------------------------------
            first_point = points[0]
            last_point = points[-1]
            
            # Get offsets from dictionary
            offsets = self.offsets[branch.Index]
            # The number of links on the branch
            nlinks = len(offsets) - 1
            
            # also get the offsets of the edge nodes, halfway the segments
            edge_offsets = [(offsets[i]+offsets[i+1])/2. for i in range(np.max([1,len(offsets)-1]))]
            
            # Check if the first and last point of the branch are already in the set            
            if (first_point not in nodes):
                first_present = False
                nodes.append(first_point)
            else:
                first_present = True
                offsets = offsets[1:]
                
            if (last_point not in nodes):
                last_present = False
                nodes.append(last_point)
            else:
                last_present = True
                offsets = offsets[:-1]
            
            # If no points remain, add an extra halfway: each branch should have at least 1 node
            if len(offsets) == 0:
                offsets = np.array([branch.geometry.length / 2.])
                edge_offsets = np.array([i*branch.geometry.length for i in [0.25,0.75]])
                nlinks += 1
                                
            # Get the index of the first and last node in the dictionary (1 based, so +1)
            i_from = nodes.index(first_point) + 1
            i_to = nodes.index(last_point) + 1
            if i_from == i_to:
                raise ValueError('Start and end node are the same. Ring geometries are not accepted.')
            network_edge_nodes.append([i_from, i_to])

            # Mesh1d edge node administration
            # -------------------------------
            # First determine the start index. This is equal to the number of already present points (+1, since 1 based)
            start_index = len(mesh1d_node_branchidx) + 1
            # For each link, create a new edge node connection
            if first_present:
                start_index -= 1
            new_edge_nodes = [[start_index + i, start_index + i + 1] for i in range(nlinks)]
            # If the first node is present, change the first point of the first edge to the existing point
            if first_present:
                new_edge_nodes[0][0] = edge_nodes_dict[first_point]
            else:
                edge_nodes_dict[first_point] = new_edge_nodes[0][0]
            # If the last node is present, change the last point of the last edge too
            if last_present:
                new_edge_nodes[-1][1] = edge_nodes_dict[last_point]
            else:
                edge_nodes_dict[last_point] = new_edge_nodes[-1][1]
            # Add to edge_nodes
            mesh1d_edge_nodes.extend(new_edge_nodes)
            
            # Update number of nodes
            mesh_point_names = [f'{branch.Index}_{offset:.2f}' for offset in offsets]
            
            # Append ids, longnames, branch and offset
            self.mesh1d.description1d['mesh1d_node_ids'].extend(mesh_point_names)
            self.mesh1d.description1d['mesh1d_node_long_names'].extend(mesh_point_names)
            mesh1d_node_branchidx.extend([i_branch + 1] * len(offsets))
            mesh1d_node_branchoffset.extend(offsets.tolist())
          
            lengths  = np.r_[0.,np.cumsum(np.hypot(np.diff(coords[i_branch],axis=0)[:,0], np.diff(coords[i_branch],axis=0)[:,1]))]
            edge_x = []
            edge_y = []
            for i_edge, edge in enumerate(edge_offsets):
                closest = (np.abs(lengths-edge)).argmin()
                if lengths[closest] > edge:
                    closest = closest - 1
                edge_x.append(edge/(lengths[closest+1]-lengths[closest])*(coords[i_branch][closest+1][0]-coords[i_branch][closest][0])+coords[i_branch][closest][0])                
                edge_y.append(edge/(lengths[closest+1]-lengths[closest])*(coords[i_branch][closest+1][1]-coords[i_branch][closest][1])+coords[i_branch][closest][1])             
               
            mesh1d_edge_branchidx.extend([i_branch+1] * len(edge_offsets))
            mesh1d_edge_branchoffset.extend(edge_offsets)
            mesh1d_edge_x.extend(edge_x)
            mesh1d_edge_y.extend(edge_y)
            
        # Parse nodes
        dimensions.nnodes = len(nodes)
        nodex, nodey = list(zip(*nodes))[:2]
        self.mesh1d.set_values('nnodex', nodex)
        self.mesh1d.set_values('nnodey', nodey)
        self.mesh1d.description1d['network_node_ids'].extend([f'{xy[0]:.0f}_{xy[1]:.0f}' for xy in nodes])
        self.mesh1d.description1d["network_node_long_names"].extend([f'x={xy[0]:.0f}_y={xy[1]:.0f}' for xy in nodes])

        # Add edge node data to mesh
        self.mesh1d.set_values('nedge_nodes', np.ravel(network_edge_nodes))
        self.mesh1d.meshgeomdim.numedge = len(mesh1d_edge_nodes)
        self.mesh1d.set_values('edge_nodes', np.ravel(mesh1d_edge_nodes))
        
        self.mesh1d.edge_branchidx  = mesh1d_edge_branchidx
        self.mesh1d.edge_branchoffset  = mesh1d_edge_branchoffset
        self.mesh1d.edge_x = mesh1d_edge_x
        self.mesh1d.edge_y = mesh1d_edge_y
                
        # Add mesh branchidx and offset to mesh
        dimensions.numnode = len(mesh1d_node_branchidx)
        self.mesh1d.set_values('branchidx', mesh1d_node_branchidx)
        self.mesh1d.set_values('branchoffsets', mesh1d_node_branchoffset)

        # Process the 1d network (determine x and y locations) and determine schematised branches
        schematised, _ = self.mesh1d.process_1d_network()
        for idx, geometry in schematised.items():
            self.schematised.at[idx, 'geometry'] = geometry
            
    def _generate_1d_spacing(self, anchor_pts, one_d_mesh_distance):
        """
        Generates 1d distances, called by function generate offsets
        """
        offsets = []
        for i in range(len(anchor_pts) - 1):
            section_length = anchor_pts[i+1] - anchor_pts[i]
            if section_length <= 0.0:
                raise ValueError('Section length must be larger than 0.0')
            nnodes = max(2, int(round(section_length / one_d_mesh_distance) + 1))
            offsets.extend(np.linspace(anchor_pts[i], anchor_pts[i+1], nnodes - 1, endpoint=False).tolist())
        offsets.append(anchor_pts[-1])

        return np.asarray(offsets)

    def generate_offsets(self, one_d_mesh_distance, structures=None, max_dist_to_struc=None):
        """
        Method to generate 1d network grid point locations. The distances are generated
        based on the 1d mesh distance and anchor points. The anchor points can for
        example be structures; every structure should be seperated by a gridpoint.
        """
        # For each branch
        for branch in self.branches.itertuples():
            # Distribute points along network [1d mesh]
            offsets = self._generate_1d_spacing([0.0, branch.geometry.length], one_d_mesh_distance)
            self.offsets[branch.Index] = offsets
        
        if structures is not None:
            # Check argument
            checks.check_argument(structures, 'structures', (pd.DataFrame, gpd.GeoDataFrame), columns=['branchid', 'chainage'])

            # Get structure data from dfs
            ids_offsets = structures[['branchid', 'chainage']]
            idx = (structures['branchid'] != '')
            if idx.any():
                logger.warning('Some structures are not linked to a branch.')
            ids_offsets = ids_offsets.loc[idx, :]

            # For each branch
            for branch_id, group in ids_offsets.groupby('branchid'):

                # Check if structures are located at the same offset
                u, c = np.unique(group['chainage'], return_counts=True)
                if any(c > 1):
                    logger.warning('Structures {} have the same location.'.format(
                        ', '.join(group.loc[np.isin(group['chainage'], u[c>1])].index.tolist())))
                
                branch = self.branches.at[branch_id, 'geometry']
                # Limits are the lengths at which the structures are located
                limits = sorted(group['chainage'].unique())
                
                anchor_pts = [0.0, branch.length]
                offsets = self._generate_1d_spacing(anchor_pts, one_d_mesh_distance)

                # Merge limits with start and end of branch
                limits = [-1e-3] + limits + [branch.length + 1e-3]
                    
                # If any structures
                if len(limits) > 2:

                    # also check if the calculation point are close enough to the structures
                    if max_dist_to_struc is not None:          
                        additional = []

                        # Skip the first and the last, these are no structures
                        for i in range(1, len(limits)-1):
                            # if the distance between two limits is large than twice the max distance to structure,
                            # the mesh point will be too far away. Add a limit on the minimum of half the length and
                            # two times the max distance
                            dist_to_prev_limit = limits[i] - (max(additional[-1], limits[i-1]) if any(additional) else limits[i-1])
                            if dist_to_prev_limit > 2 * max_dist_to_struc:
                                additional.append(limits[i] - min(2 * max_dist_to_struc, dist_to_prev_limit / 2))

                            dist_to_next_limit = limits[i+1] - limits[i]
                            if dist_to_next_limit > 2 * max_dist_to_struc:
                                additional.append(limits[i] + min(2 * max_dist_to_struc, dist_to_next_limit / 2))

                        # Join the limits
                        limits = sorted(limits + additional)
                          
                    # Get upper and lower limits
                    upper_limits = limits[1:]
                    lower_limits = limits[:-1]
                    
                    # Determine the segments that are missing a grid point
                    in_range = [((offsets > lower) & (offsets < upper)).any() for lower, upper in zip(lower_limits, upper_limits)]

                    while not all(in_range):
                        # Get the index of the first segment without grid point
                        i = in_range.index(False)
                        
                        # Add it to the anchor pts
                        anchor_pts.append((lower_limits[i] + upper_limits[i]) / 2.)
                        anchor_pts = sorted(anchor_pts)
                        
                        # Generate new offsets
                        offsets = self._generate_1d_spacing(anchor_pts, one_d_mesh_distance)
                
                        # Determine the segments that are missing a grid point
                        in_range = [((offsets > lower) & (offsets < upper)).any() for lower, upper in zip(lower_limits, upper_limits)]
                    
                    if len(anchor_pts) > 2:
                        logger.info(f'Added 1d mesh nodes on branch {branch_id} at: {anchor_pts}, due to the structures at {limits}.')

                # Set offsets for branch id
                self.offsets[branch_id] = offsets

    def get_roughness_description(self, roughnesstype, value):

        if np.isnan(value):
            raise ValueError('Roughness value should not be NaN.')

        # Check input
        checks.check_argument(roughnesstype, 'roughness type', (str, int))
        checks.check_argument(value, 'roughness value', (float, int, np.float, np.integer))

        # Convert integer to string
        if isinstance(roughnesstype, int):
            roughnesstype = hydamo_to_dflowfm.roughness_gml[roughnesstype]
        
        # Get name
        name = f'{roughnesstype}_{float(value)}'

        # Check if the description is already known
        if name.lower() in map(str.lower, self.roughness_definitions.keys()):
            return name

        # Convert roughness type string to integer for dflowfm
        delft3dfmtype = roughnesstype

        if roughnesstype.lower() == 'stricklerks':
            raise ValueError()

        # Add to dict
        self.roughness_definitions[name] = {
            'name': name,
            'code': delft3dfmtype,
            'value': value
        }

        return name

    def add_mesh2d(self, twodmesh):
        """
        Add 2d mesh to network object.
        """
        if not hasattr(twodmesh, 'meshgeom'):
            checks.check_argument(twodmesh, 'twodmesh', meshgeom)
            geometries = twodmesh
        else:
            if not isinstance(twodmesh.meshgeom, meshgeom):
                raise TypeError('The given mesh should have an attribute "meshgeom".')
            geometries = twodmesh.meshgeom

        # Add the meshgeom
        self.mesh2d.add_from_other(geometries)

        # Add mdu parameters
        if hasattr(twodmesh, 'missing_z_value'):
            if twodmesh.missing_z_value is not None:
                self.mdu_parameters['BedlevUni'] = twodmesh.missing_z_value

    def get_node_idx_offset(self, branch_id, pt, nnodes=1):
        """
        Get the index and offset of a node on a 1d branch.
        The nearest node is looked for.
        """

        # Project the point on the branch
        dist = self.schematised[branch_id].project(pt)

        # Get the branch data from the networkdata
        branchidx = self.mesh1d.description1d['network_branch_ids'].index(self.str2chars(branch_id, self.idstrlength)) + 1
        pt_branch_id = self.mesh1d.get_values('branchidx', as_array=True)
        idx = np.where(pt_branch_id == branchidx)
        
        # Find nearest offset
        offsets = self.mesh1d.get_values('branchoffset', as_array=True)[idx]
        isorted = np.argsort(np.absolute(offsets - dist))
        isorted = isorted[:min(nnodes, len(isorted))]

        # Get the offset
        offset = [offsets[imin] for imin in isorted]
        # Get the id of the node
        node_id = [idx[0][imin] + 1 for imin in isorted]

        return node_id, offset


class Structures:

    def __init__(self, dflowfmmodel):
        self.pumps = {}
        self.weirs = {}
        self.bridges = {}
        self.uweirs= {}
        self.culverts = {}
        self.orifices = {}
        self.compounds = {}
        
        self.dflowfmmodel = dflowfmmodel

        # Create the io class
        self.io = dfmreader.StructuresIO(self)

    def add_pump(self, id, branchid, chainage, orientation, numstages, controlside, capacity, startlevelsuctionside, stoplevelsuctionside):
        self.pumps[id] = {
            "type": "pump",
            'id': id,
            'branchid': branchid,
            'chainage': chainage,
            'orientation': orientation,
            'numstages': numstages,
            'controlSide': controlside,
            'capacity': capacity,
            'startlevelSuctionSide': startlevelsuctionside,
            'stoplevelSuctionSide': stoplevelsuctionside            
        }

    def add_weir(self, id, branchid, chainage, crestlevel, crestwidth, allowedflowdir='both', corrcoeff=1.0, usevelocityheight='true'):
        self.weirs[id] = {
            "type": "weir",
            'id': id,
            'branchid': branchid,
            'chainage': chainage,
            'allowedFlowDir': allowedflowdir,
            'crestLevel': crestlevel,
            'crestWidth': crestwidth,
            'corrCoeff': corrcoeff,
            'useVelocityHeight': usevelocityheight            
        }

    def add_orifice(self, id, branchid, chainage, crestlevel, crestwidth, gateloweredgelevel,                                        allowedflowdir='both', uselimitflowpos=False, limitflowpos=0.0, uselimitflowneg=False, limitflowneg=0.0, corrcoeff=1.0, usevelocityheight='true'):
        self.orifices[id] = {
            "type": "orifice",
            'id': id,
            'branchid': branchid,
            'chainage': chainage,
            'allowedFlowDir': allowedflowdir,
            'crestLevel': crestlevel,
            'crestWidth': crestwidth,
            'gateLowerEdgeLevel': gateloweredgelevel,            
            'useLimitFlowPos': uselimitflowpos,
            'limitFlowPos': limitflowpos,
            'useLimitFlowNeg': uselimitflowneg,
            'limitFlowNeg': limitflowneg,
            'corrCoeff': corrcoeff,
            'useVelocityHeight': usevelocityheight            
        }
    
    def add_bridge(self, id, branchid, chainage, length, bedlevel, upperheight, lowerheight,crosssection,
                   inletlosscoeff,outletlosscoeff,allowedflowdir='both',
                   frictiontype='Strickler', frictionvalue=75.0):
        """
        Add a bridge to the schematisation.

        Note that the cross section should be handed as dictionary. This should contain the
        shape (circle, rectangle) and the required arguments.
        
        For now, we use the smae cross sections as culverts. This needs to be refined.       
        
        """     
         
        
        # Check the content of the cross section dictionary
        #checks.check_dictionary(crosssection, required='shape', choice=['diameter', ['width', 'height', 'closed']])

        self.bridges[id] = {
            "type": "bridge",
            "id": id,
            "branchid": branchid,
            "chainage": chainage,            
            "allowedFlowDir": allowedflowdir,
            "csDefId": crosssection,
            "bedLevel": bedlevel,
            "inletLossCoeff": inletlosscoeff,
            "outletLossCoeff": outletlosscoeff,            
            "frictionType": frictiontype,
            "friction": frictionvalue,
            "length":length
        }
        
    def add_uweir(self, id, branchid, chainage, crestlevel, numlevels, yvalues, zvalues, allowedflowdir='both', dischargecoeff=1.0):
        
        """
        Add a universal weir to the schematisation.

                
        """            
        
        self.uweirs[id] = {
            "type": "universalWeir",
            "id": id,
            "branchid": branchid,
            "chainage": chainage,            
            "allowedFlowDir": allowedflowdir,            
            "numLevels": numlevels,            
            "yValues": yvalues,
            "zValues": zvalues,            
            "crestLevel": crestlevel,
            "dischargeCoeff": dischargecoeff                        
        }
        
    def add_culvert(self, id, branchid, chainage, leftlevel, rightlevel, crosssection, length, inletlosscoeff,
                    outletlosscoeff, allowedflowdir='both', 
                    valveonoff=0, numlosscoeff=0, valveopeningheight=None, relopening=None, losscoeff=None,
                    frictiontype='Strickler', frictionvalue=75.0):
        """
        Add a culvert to the schematisation.

        Note that the cross section should be handed as dictionary. This should contain the
        shape (circle, rectangle) and the required arguments.
        
        """

        # Check the content of the cross section dictionary
        checks.check_dictionary(crosssection, required='shape', choice=['diameter', ['width', 'height', 'closed']])

        # Get roughnessname
        roughnessname = self.dflowfmmodel.network.get_roughness_description(frictiontype, frictionvalue)

        # Add cross section definition
        # WORKAROUND: for the GUI, a profile definition has to be created for every culvert. We do this temporary.
        if crosssection['shape'] == 'circle':
            definition = self.dflowfmmodel.crosssections.add_circle_definition(crosssection['diameter'], frictiontype, frictionvalue, name=id)
        elif crosssection['shape'] == 'rectangle':
            definition = self.dflowfmmodel.crosssections.add_rectangle_definition(
                crosssection['height'], crosssection['width'], crosssection['closed'], frictiontype, frictionvalue, name=id)
        else:
            NotImplementedError(f'Cross section with shape \"{crosssection["shape"]}\" not implemented.')

        # Add the culvert to the dictionary
        self.culverts[id] = {
            "type": "culvert",
            "id": id,
            "branchid": branchid,
            "chainage": chainage,
            "allowedFlowDir": allowedflowdir,
            "leftLevel": leftlevel,
            "rightLevel": rightlevel,
            "csDefId": definition,
            "length": round(length, 3),
            "inletLossCoeff": inletlosscoeff,
            "outletLossCoeff": outletlosscoeff,
            "valveOnOff": int(valveonoff),                        
            "numLossCoeff": int(numlosscoeff),            
            "bedFrictionType": frictiontype,
            "bedFriction": frictionvalue            
        }
        if valveonoff > 0:
            self.culverts[id]['valveOpeningHeight'] = valveopeningheight
            self.culverts[id]['relOpening']  =  relopening
            self.culverts[id]['lossCoeff']  = losscoeff
            
    def add_compound(self, id, numstructures, structurelist):
        self.compounds[id] = {
            "type": "compound",
            "id": id,
            "numStructures": numstructures,
            "structureIds": structurelist            
        }

    def as_dataframe(self, pumps=False, weirs=False, bridges=False, culverts=False, uweirs=False, orifices=False, compounds=False):
        """
        Returns a dataframe with the structures. Specify with the keyword arguments what structure types need to be returned.
        """
        dfs = []
        for df, descr, add in zip([self.culverts, self.weirs, self.bridges, self.pumps, self.uweirs, self.orifices, self.compounds], ['culvert', 'weir','bridge', 'pump', 'uweir','orifice','compound'], [culverts, weirs, bridges, pumps, uweirs, orifices, compounds]):
            if any(df) and add:
                df = pd.DataFrame.from_dict(df, orient='index')
                df.insert(loc=0, column='structype', value=descr, allow_duplicates=True)
                dfs.append(df)

        if len(dfs) > 0:
            return pd.concat(dfs, sort=False)
        
class ObservationPoints(ExtendedGeoDataFrame):

    def __init__(self, dflowfmmodel):
        super(ObservationPoints, self).__init__(geotype=Point, required_columns=['name', 'branchId', 'chainage', 'geometry', 'locationType'])

        self._metadata.append('dflowfmmodel')
        self.dflowfmmodel = dflowfmmodel
          

    def add_points(self, crds, names, locationTypes=None, snap_distance=None):
        """
        Method to add observation points to schematisation. Observation points can be of type '1d' or '2d'. 1d-points are snapped to the branch.

        Parameters
        ----------
        crds : Nx2 list or array
            x and y coordinates of observation points
        names : str or list
            names of the observation points
        locationTypes:  str or list
            type of the observationpoints: 1d or 2d
        """
        if snap_distance is None:
            snap_distance = 5
        if isinstance(names,str):
            names = [names]
            crds = [crds]
        if locationTypes is not None:
            if isinstance(names,str):
                locationTypes = [locationTypes]
            # split 1d and 2d points, as the first ones need to be snapped to branches
            obs2d = gpd.GeoDataFrame()
            obs2d['name'] = [n for nn,n in enumerate(names) if locationTypes[nn]=='2d']
            obs2d['locationType'] = '2d'
            obs2d['geometry'] = [Point(*pt) for ipt,pt in enumerate(crds) if (locationTypes[ipt]=='2d')&(not isinstance(pt, Point))]
            obs2d['x'] = [pt.coords[0][0] for pt in obs2d['geometry']]
            obs2d['y'] = [pt.coords[0][1] for pt in obs2d['geometry']]
            names1d = [n for n_i,n in enumerate(names) if locationTypes[n_i]=='1d']
            crds1d = [c for c_i,c in enumerate(crds) if locationTypes[c_i]=='1d']
        else:
            names1d = names
            crds1d = crds

            
        # Check if data for snapping is available
        network = self.dflowfmmodel.network
        if not network.mesh1d.meshgeomdim.numnode:
            raise ValueError('The network geometry should be generated before the observation points can be snapped to 1d.')
        
        obs1d = gpd.GeoDataFrame()
        obs1d['name'] = names1d
        obs1d['geometry'] = [Point(*pt) if not isinstance(pt, Point) else pt for pt in crds1d]        
        obs1d['locationType'] = '1d'        
        geometry.find_nearest_branch(network.branches, obs1d, method='overal', maxdist=snap_distance)      
        obs1d.rename(columns={'branch_id':'branchId', 'branch_offset': 'chainage'}, inplace=True)
        obs = obs1d.append(obs2d, sort=True) if locationTypes is not None else obs1d
                    
        # Add to dataframe        
        self.set_data(obs, index_col="name", check_columns=True)
                    
def list_to_str(lst):
    string = ' '.join([f'{number:6.3f}' for number in lst])
    return string



# coding: utf-8
import os, sys, math, logging
from netCDF4 import Dataset
from collections import OrderedDict
from datetime import *
import numpy as np


class UgridWriter:
    """Writer for FM files"""
    
    def __init__(self):
        self.idstrlength = 40
        self.longstrlength = 80

    def write(self, dflowfmmodel, path, version):  # write ugrid file from GWSW model

        ncfile = self.create_netcdf(path, version)

        # Write the 1d mesh
        if not dflowfmmodel.network.mesh1d.empty():
            self.init_1dnetwork(ncfile, dflowfmmodel.network.mesh1d)
            self.set_1dmesh(ncfile, dflowfmmodel.network.mesh1d)
            self.set_1dnetwork(ncfile, dflowfmmodel.network.mesh1d)

        # Write the 2d mesh
        if not dflowfmmodel.network.mesh2d.empty():
            self.init_2dmesh(ncfile, dflowfmmodel.network.mesh2d)
            self.set_2dmesh(ncfile, dflowfmmodel.network.mesh2d)

        # Write the 1d 2d links
        if dflowfmmodel.network.links1d2d.nodes1d:
            self.init_1d2dlinks(ncfile, dflowfmmodel.network.links1d2d)
            self.set_1d2dlinks(ncfile, dflowfmmodel.network.links1d2d)

        ncfile.close()

    @staticmethod
    def to_char_list(lst, size):
        """Convert list of strings to list of stings with a fixed number of characters"""
        return [item.ljust(size)[:size] for item in lst]

    def create_netcdf(self, path, version):

        # File format:
        outformat = "NETCDF3_CLASSIC" #"NETCDF4"
        # File where we going to write
        ncfile = Dataset(path, 'w', format=outformat)

        # global attributes
        ncfile.Conventions = "CF-1.8 UGRID-1.0"
        ncfile.title = 'Delft3D-FM 1D2D network for model '+os.path.split(path)[-1].rstrip('_net.nc')
        ncfile.source = f"delft3dfmpy v.{version['number']}, D-HyDAMO, model {os.path.split(path)[-1].rstrip('_net.nc')}"
        ncfile.history = f"Created on {version['date']} by {os.path.split(__file__)[-1]}."        
        ncfile.institution = "Deltares/HKV"
        ncfile.references = "https://github.com/openearth/delft3dfmpy/; https://www.deltares.nl; https://www.hkv.nl"      
        ncfile.comment = f"Tested and compatible with D-Flow FM {version['dfm_version']}, DIMRset {version['dimr_version']} and D-HYDRO suite 1D2D {version['suite_version']}"

        return ncfile

    def init_1dnetwork(self, ncfile, mesh1d):

        # dimensions of the network
        ncfile.createDimension("time", None)
        ncfile.createDimension("network1d_nEdges", mesh1d.meshgeomdim.nbranches)
        ncfile.createDimension("network1d_nNodes", mesh1d.meshgeomdim.nnodes)
        ncfile.createDimension("network1d_nGeometryNodes", mesh1d.meshgeomdim.ngeometry)        
        ncfile.createDimension("idstrlength", self.idstrlength)
        ncfile.createDimension("longstrlength", self.longstrlength)
        ncfile.createDimension("mesh1d_nEdges", mesh1d.meshgeomdim.numedge)
        ncfile.createDimension("mesh1d_nNodes", mesh1d.meshgeomdim.numnode)
        ncfile.createDimension("Two", 2)

    def init_2dmesh(self, ncfile, cmesh2d):

        # Create dimensions
        ncfile.createDimension("max_nmesh2d_face_nodes", cmesh2d.meshgeomdim.maxnumfacenodes)
        ncfile.createDimension("mesh2d_nEdges", cmesh2d.meshgeomdim.numedge)
        ncfile.createDimension("mesh2d_nFaces", cmesh2d.meshgeomdim.numface)
        ncfile.createDimension("mesh2d_nNodes", cmesh2d.meshgeomdim.numnode)

    def init_1d2dlinks(self, ncfile, links1d2d):
        # Dimensions
        ncfile.createDimension("nLink1D2D_edge", len(links1d2d.nodes1d))
        
        cm = ncfile.createVariable("composite_mesh", "i4", ())
        cm.cf_role = 'parent_mesh_topology'
        cm.meshes= 'mesh1D mesh2D'
        cm.mesh_contact = 'link1d2d'

    def set_1dnetwork(self, ncfile, cmesh1d):

        #  network topology
        ntw = ncfile.createVariable("network1d", "i4", ())
        ntw.cf_role = 'mesh_topology'
        ntw.edge_dimension = 'network1d_nEdges'
        ntw.edge_geometry = 'network1d_geometry'
        ntw.edge_node_connectivity = 'network1d_edge_nodes'
        ntw.long_name = "Topology data of 1D network"
        ntw.node_coordinates = 'network1d_node_x network1d_node_y'
        ntw.node_dimension = 'network1d_nNodes'
        ntw.topology_dimension = 1        
        ntw.node_id = "network1d_node_id"
        ntw.node_long_name = "network1d_node_long_name"
        ntw.branch_id = "network1d_branch_id"
        ntw.branch_long_name = "network1d_branch_long_name"
        ntw.edge_length = "network1d_edge_length"
        ntw.branch_order = "network1d_branch_order"

        ntw_node_id = ncfile.createVariable("network1d_node_id", "c", ("network1d_nNodes", "idstrlength"))
        ntw_node_id.long_name = "ID of network nodes"        
        ntw_node_id[:] = self.to_char_list(cmesh1d.description1d["network_node_ids"], self.idstrlength)

        ntw_node_longname = ncfile.createVariable("network1d_node_long_name", "c", ("network1d_nNodes", "longstrlength"))
        ntw_node_longname.long_name = "Long name of network nodes"
        ntw_node_longname[:] = self.to_char_list(cmesh1d.description1d["network_node_long_names"], self.longstrlength)

        # network nodes
        ntw_node_x = ncfile.createVariable("network1d_node_x", np.float64, "network1d_nNodes")
        ntw_node_x.standard_name = 'projection_x_coordinate'
        ntw_node_x.long_name = "x coordinates of network nodes"
        ntw_node_x.units = 'm'
        ntw_node_x[:] = cmesh1d.get_values('nnodex', as_array=True)

        ntw_node_y = ncfile.createVariable("network1d_node_y", np.float64, "network1d_nNodes")
        ntw_node_y.standard_name = 'projection_y_coordinate'
        ntw_node_y.long_name = "y coordinates of network nodes"
        ntw_node_y.units = 'm'
        ntw_node_y[:] = cmesh1d.get_values('nnodey', as_array=True)

        ntw_branch_id_name = ncfile.createVariable("network1d_branch_id", "c", ("network1d_nEdges", "idstrlength"))
        ntw_branch_id_name.long_name = "ID of branch geometries"
        ntw_branch_id_name[:] = self.to_char_list(cmesh1d.description1d["network_branch_ids"], self.idstrlength)

        ntw_branch_id_longname = ncfile.createVariable("network1d_branch_long_name", "c", ("network1d_nEdges", "longstrlength"))
        ntw_branch_id_longname.long_name = "Long name of branch geometries"
        ntw_branch_id_longname[:] = self.to_char_list(cmesh1d.description1d["network_branch_long_names"], self.longstrlength)

        ntw_edge_length = ncfile.createVariable("network1d_edge_length", np.float64, "network1d_nEdges")        
        ntw_edge_length.long_name = "Real length of branch geometries"
        ntw_edge_length.units = "m"
        ntw_edge_length[:] = cmesh1d.get_values('nbranchlengths', as_array=True)

        ntw_branch_order = ncfile.createVariable("network1d_branch_order", "i4", "network1d_nEdges")        
        ntw_branch_order.long_name = "Order of branches for interpolation"
        ntw_branch_order.mesh = "network1d"
        ntw_branch_order.location = "edge"
        ntw_branch_order[:] = cmesh1d.get_values('nbranchorder', as_array=True)

        # network edges
        ntw_edge_node = ncfile.createVariable("network1d_edge_nodes", "i4", ("network1d_nEdges", "Two"))
        ntw_edge_node.cf_role = 'edge_node_connectivity'
        ntw_edge_node.long_name = 'start and end nodes of network edges'
        ntw_edge_node.start_index = 1
        ntw_edge_node[:] = cmesh1d.get_values('nedge_nodes', as_array=True)

        # network geometry
        ntw_geom = ncfile.createVariable("network1d_geometry", "i4", ())
        ntw_geom.geometry_type = 'line'
        ntw_geom.long_name = "1D Geometry"
        ntw_geom.node_count = "network1d_geom_node_count"     
        ntw_geom.node_coordinates = 'network1d_geom_x network1d_geom_y'
        
        ntw_geom_node_count = ncfile.createVariable("network1d_geom_node_count", "i4", "network1d_nEdges")
        ntw_geom_node_count.long_name = "Number of geometry nodes per branch"
        ntw_geom_node_count[:] = cmesh1d.get_values('nbranchgeometrynodes', as_array=True)

        ntw_geom_x = ncfile.createVariable("network1d_geom_x", np.float64, ("network1d_nGeometryNodes"))
        ntw_geom_x.standard_name = 'projection_x_coordinate'
        ntw_geom_x.long_name = 'x-coordinate of branch geometry nodes'
        ntw_geom_x.units = 'm' 
        ntw_geom_x[:] = cmesh1d.get_values('ngeopointx', as_array=True)        
        
        ntw_geom_y = ncfile.createVariable("network1d_geom_y", np.float64, ("network1d_nGeometryNodes"))
        ntw_geom_y.standard_name = 'projection_y_coordinate'
        ntw_geom_y.long_name = 'y-coordinate of branch geometry nodes'
        ntw_geom_y.units = 'm'           
        ntw_geom_y[:] = cmesh1d.get_values('ngeopointy', as_array=True)

    def set_1dmesh(self, ncfile, cmesh1d):

        mesh1d = ncfile.createVariable("mesh1d", "i4", ())
        mesh1d.cf_role = 'mesh_topology'
        mesh1d.long_name = "Topology data of 1D Mesh"
        mesh1d.coordinate_space = 'network1d'
        mesh1d.edge_dimension = 'mesh1d_nEdges'
        mesh1d.edge_node_connectivity = 'mesh1d_edge_nodes'
        mesh1d.edge_coordinates = 'mesh1d_edge_branch mesh1d_edge_offset mesh1d_edge_x mesh1d_edge_y'        
        mesh1d.node_coordinates = 'mesh1d_node_branch mesh1d_node_offset'
        mesh1d.node_dimension = 'mesh1d_nNodes'
        mesh1d.node_id = "mesh1d_node_id"
        mesh1d.node_long_name = "mesh1d_node_long_name"
        mesh1d.topology_dimension = 1

        mesh1d_node_id = ncfile.createVariable("mesh1d_node_id", "c", ("mesh1d_nNodes", "idstrlength"))        
        mesh1d_node_id.long_name = "ID of mesh nodes"        
        mesh1d_node_id[:] = self.to_char_list(cmesh1d.description1d["mesh1d_node_ids"], self.idstrlength)

        mesh1d_node_longname = ncfile.createVariable("mesh1d_node_long_name", "c", ("mesh1d_nNodes", "longstrlength"))
        mesh1d_node_longname.long_name = 'Long name of mesh nodes'                
        mesh1d_node_longname[:] = self.to_char_list(cmesh1d.description1d["mesh1d_node_long_names"], self.longstrlength)

        mesh1d_edge_node = ncfile.createVariable("mesh1d_edge_nodes", "i4", ("mesh1d_nEdges", "Two"))
        mesh1d_edge_node.cf_role  = "edge_node_connectivity"
        mesh1d_edge_node.long_name = 'Start and end nodes of mesh edges'
        mesh1d_edge_node.start_index = 1       
        mesh1d_edge_node[:] = cmesh1d.get_values("edge_nodes", as_array=True) 

        mesh1d_edge_branch = ncfile.createVariable("mesh1d_edge_branch", "i4", "mesh1d_nEdges")
        mesh1d_edge_branch.long_name = "Index of branch on which mesh edges are located"
        mesh1d_edge_branch.start_index = 1
        mesh1d_edge_branch[:] = np.ravel(cmesh1d.edge_branchidx)           
       
        mesh1d_edge_offset = ncfile.createVariable("mesh1d_edge_offset", np.float64, "mesh1d_nEdges")
        mesh1d_edge_offset.long_name = "Offset along branch of mesh edges"
        mesh1d_edge_offset.units = "m"        
        mesh1d_edge_offset[:] = np.ravel(cmesh1d.edge_branchoffset)
        
        mesh1d_edge_x = ncfile.createVariable("mesh1d_edge_x", np.float64, "mesh1d_nEdges")
        mesh1d_edge_x.long_name = "x-coordinate of mesh edges"
        mesh1d_edge_x.units = "m"        
        mesh1d_edge_x[:] = np.ravel(cmesh1d.edge_x)
        
        mesh1d_edge_y = ncfile.createVariable("mesh1d_edge_y", np.float64, "mesh1d_nEdges")
        mesh1d_edge_y.long_name = "y-coordinate of mesh edges"
        mesh1d_edge_y.units = "m"        
        mesh1d_edge_y[:] = np.ravel(cmesh1d.edge_y)
        
        mesh1d_node_branch = ncfile.createVariable("mesh1d_node_branch", "i4", "mesh1d_nNodes")
        mesh1d_node_branch.long_name = "Index of branch on which mesh nodes are located"
        mesh1d_node_branch.start_index = 1
        mesh1d_node_branch[:] = cmesh1d.get_values("branchidx", as_array=True)

        mesh1d_node_offset = ncfile.createVariable("mesh1d_node_offset", np.float64, "mesh1d_nNodes", fill_value= 0.)        
        mesh1d_node_offset.long_name = "Offset along branch of mesh nodes"
        mesh1d_node_offset.units = "m"
        mesh1d_node_offset[:] = cmesh1d.get_values("branchoffsets", as_array=True)
       
    # set 2d mesh data to netcdf file
    def set_2dmesh(self, ncfile, cmesh2d):

        mesh2d = ncfile.createVariable("mesh2d", "i4", ())
        mesh2d.long_name = "Topology data of 2D network"
        mesh2d.topology_dimension = 2
        mesh2d.cf_role = 'mesh_topology'
        mesh2d.node_coordinates = 'mesh2d_node_x mesh2d_node_y'
        mesh2d.node_dimension = 'mesh2d_nNodes'
        mesh2d.edge_coordinates = 'mesh2d_edge_x mesh2d_edge_y'
        mesh2d.edge_dimension = 'mesh2d_nEdges'
        mesh2d.edge_node_connectivity = 'mesh2d_edge_nodes'
        mesh2d.face_node_connectivity = 'mesh2d_face_nodes'
        mesh2d.max_face_nodes_dimension = 'max_nmesh2d_face_nodes'
        mesh2d.face_dimension = "mesh2d_nFaces"
        #mesh2d.edge_face_connectivity = "mesh2d_edge_faces"
        mesh2d.face_coordinates = "mesh2d_face_x mesh2d_face_y"

        # Nodes:
        mesh2d_x = ncfile.createVariable("mesh2d_node_x", np.float64, mesh2d.node_dimension)
        mesh2d_y = ncfile.createVariable("mesh2d_node_y", np.float64, mesh2d.node_dimension)
        mesh2d_z = ncfile.createVariable("mesh2d_node_z", np.float64, mesh2d.node_dimension, fill_value=-999.0)

        mesh2d_x.standard_name = 'projection_x_coordinate'
        mesh2d_y.standard_name = 'projection_y_coordinate'
        mesh2d_z.standard_name = 'altitude'

        for var, dim in zip([mesh2d_x, mesh2d_y, mesh2d_z],  list('xyz')):
            setattr(var, 'units', 'm')
            setattr(var, 'mesh', 'mesh2d')
            setattr(var, 'location', 'node')
            setattr(var, 'long_name', f'{dim}-coordinate of mesh nodes')

        mesh2d_z.coordinates = 'mesh2d_node_x mesh2d_node_y'
        mesh2d_z.grid_mapping = ''
        
        mesh2d_x[:] = cmesh2d.get_values("nodex")
        mesh2d_y[:] = cmesh2d.get_values("nodey")
        
        # Edges:
        # mesh2d_xu = ncfile.createVariable("mesh2d_edge_x", np.float64,  "nmesh2d_edges")
        # mesh2d_yu = ncfile.createVariable("mesh2d_edge_y", np.float64,  "nmesh2d_edges")
        # mesh2d_xu[:] = cmesh2d.get_values("edgex")
        # mesh2d_yu[:] = cmesh2d.get_values("edgey")
        # mesh2d_xu.long_name = 'x-coordinate of mesh edges'
        # mesh2d_yu.long_name = 'y-coordinate of mesh edges'

        # for var, dim in zip([mesh2d_xu, mesh2d_yu],  list('xy')):
        #     setattr(var, 'units', 'm')
        #     setattr(var, 'mesh', 'mesh2d')
        #     setattr(var, 'location', 'edge')
        #     setattr(var, 'standard_name', f'projection_{dim}_coordinate')

        mesh2d_en = ncfile.createVariable("mesh2d_edge_nodes", "i4", ( mesh2d.edge_dimension, "Two"), fill_value=-999)
        mesh2d_en.cf_role = 'edge_node_connectivity'
        mesh2d_en.long_name = 'maps every edge to the two nodes that it connects'
        mesh2d_en.start_index = 1
        mesh2d_en.location = 'edge'
        mesh2d_en.mesh = 'mesh2d'
        mesh2d_en[:] = cmesh2d.get_values('edge_nodes', as_array=True)

        # mesh2d_et = ncfile.createVariable("mesh2d_edge_types", "i4", mesh2d.edge_dimension)
        # mesh2d_et.long_name = 'edge type (relation between edge and flow geometry)'
        # mesh2d_et.coordinates = 'mesh2d_edge_x mesh2d_edge_y'
        # mesh2d_et.location = 'edge'
        # mesh2d_et.mesh = 'mesh2d'
        # mesh2d_et.standard_name = ''
        # mesh2d_et.units = ''
        # mesh2d_et[:] = 2

        mesh2d_fn = ncfile.createVariable("mesh2d_face_nodes", "i4", (mesh2d.face_dimension, mesh2d.max_face_nodes_dimension), fill_value=-999)
        mesh2d_fn.cf_role = 'face_node_connectivity'
        mesh2d_fn.mesh = 'mesh2d'
        mesh2d_fn.location = 'face'
        mesh2d_fn.long_name = 'maps every face to the nodes that it defines'
        mesh2d_fn.start_index = 1
        mesh2d_fn[:] = cmesh2d.get_values('face_nodes', as_array=True)

        mesh2d_face_x = ncfile.createVariable("mesh2d_face_x", np.float64, mesh2d.face_dimension)
        mesh2d_face_y = ncfile.createVariable("mesh2d_face_y", np.float64, mesh2d.face_dimension)
        mesh2d_face_z = ncfile.createVariable("mesh2d_face_z", np.float64, mesh2d.face_dimension, fill_value=-999.0)

        for var, dim in zip([mesh2d_face_x, mesh2d_face_y, mesh2d_face_z],  list('xyz')):
            setattr(var, 'units', 'm')
            setattr(var, 'mesh', 'mesh2d')
            setattr(var, 'location', 'face')
            setattr(var, 'standard_name', f'projection_{dim}_coordinate' if dim != 'z' else 'altitude')
            setattr(var, 'long_name', f'{dim}-coordinate of face nodes')

        mesh2d_face_z.coordinates = 'mesh2d_face_x mesh2d_face_y'
        mesh2d_face_z.grid_mapping = ''
    
        mesh2d_face_x[:] = cmesh2d.get_values("facex")
        mesh2d_face_y[:] = cmesh2d.get_values("facey")

        # Assign altitude data
        # To faces
        if cmesh2d.is_allocated('facez'):
            mesh2d_face_z[:] = cmesh2d.get_values("facez")
        # Assign to nodes
        if cmesh2d.is_allocated('nodez'):
            mesh2d_z[:] = cmesh2d.get_values("nodez")
        # Raise error if none of both is allocated
        if not cmesh2d.is_allocated('nodez') and not cmesh2d.is_allocated('facez'):
            raise ValueError('Assign altitude values either to nodes or faces.')


    def set_1d2dlinks(self, ncfile, links1d2d):
        
        nlinks = len(links1d2d.nodes1d)

        link1d2d = ncfile.createVariable("link1d2d", "i4", ("nLink1D2D_edge", "Two"))
        link1d2d.cf_role = 'mesh_topology_contact'
        link1d2d.contact= 'mesh1D:node mesh2D:face'
        link1d2d.contact_type = 'link1d2d_contact_type'
        link1d2d.contact_ids = 'link1d2d_ids'
        link1d2d.contact_long_names = 'link1d2d_long_names'
        link1d2d.start_index = 1
        link1d2d[:, :] = np.c_[links1d2d.nodes1d, links1d2d.faces2d]

        link1d2d_ids = ncfile.createVariable("link1d2d_ids", "c", ("nLink1D2D_edge", "idstrlength"))
        link1d2d_ids.long_name = 'ids of the contact'
        link1d2d_ids[:] = [self.str2chars(f'linkid{i+1}', self.idstrlength) for i in range(nlinks)]

        link1d2d_long_names = ncfile.createVariable("link1d2d_long_names", "c", ("nLink1D2D_edge", "longstrlength"))
        link1d2d_long_names.long_name = 'long names of the contact'
        link1d2d_long_names[:] = [self.str2chars(f'longnames{i+1}', self.longstrlength) for i in range(nlinks)]

        link1d2d_contact_type = ncfile.createVariable("link1d2d_contact_type", "i4", "nLink1D2D_edge", fill_value=-1)
        link1d2d_contact_type[:] = [3] * nlinks

    def str2chars(self, string, size):
        return string.ljust(size)[:size]
    
    # def add_bounds(self, ncfile):



# coding: latin-1
import os
import re
import shutil
import datetime

import numpy as np
from shapely.geometry import box, LineString, Point
import delft3dfmpy
from delft3dfmpy.io.UgridWriter import UgridWriter
from delft3dfmpy.core.geometry import orthogonal_line

import logging

logger = logging.getLogger(__name__)

class DFlowFMWriter:
    """Writer for FM files"""

    # versioning info
    version = { 'number'        : delft3dfmpy.__version__,
                'date'          : datetime.datetime.strftime(datetime.datetime.today(),'%Y-%m-%d %H:%M:%S'),
                'dfm_version'   : '1.2.105.67228M',
                'dimr_version'  : '2.13.02.67836',
                'suite_version' : '(Beta) 0.9.6.51435'} 
   
    def __init__(self, dflowfmmodel, output_dir, name):
        self.dflowfmmodel = dflowfmmodel
        self.output_dir = os.path.join(output_dir, 'fm')
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        self.name = name
       
        self.version = DFlowFMWriter.version
        
        self.mdu_parameters = dflowfmmodel.mdu_parameters
        self.mdu_parameters['NetFile'] = self.name + "_net.nc"

        self.mdu_parameters['ExtForceFileNew'] = self.name + "_new.ext"

        self.mdufile = os.path.join(self.output_dir, self.name +'.mdu')
        self.extfile_new = os.path.join(self.output_dir, self.mdu_parameters['ExtForceFileNew'])
        self.netfile = os.path.join(self.output_dir, self.mdu_parameters['NetFile'])
        
        if hasattr(self.dflowfmmodel, 'dimr_path'):
           self.run_dimrpad = self.dflowfmmodel.dimr_path
       
        with open(os.path.join(self.output_dir, 'boundaries.bc'), 'w') as f:
            self._write_header(f, 'boundConds', 1.01, extra_linebreak=False)

    def write_all(self):  # write all fm files from HyDAMO
        """
        Wrapper to write all components to DFM. Remove existing files and create new ones. Note that the old format ext-file is not used anymore,.
        """
        if os.path.exists(self.extfile_new):
            os.remove(self.extfile_new)

        with open(self.extfile_new,'w') as f:
            self._write_header(f, 'extForce', 2.01)

        # Write grid in Ugrid format to netcdf
        ugridWriter = UgridWriter()
        ugridWriter.write(dflowfmmodel=self.dflowfmmodel, path=self.netfile, version=self.version)

        # mdu file
        self.writeMdFiles()
        # Roughness files
        self.write_roughness()
        # Cross section definitions
        self.write_crosssection_definitions()
        # Cross section locations
        self.write_crosssection_locations()
        # Structures (weirs, pumps, culverts)
        self.write_structures()
        # Initial conditions
        self.write_initial_conditions()
        # Boundary conditions
        self.write_boundary_conditions()
        # Observation points
        self.write_observation_points()
        if hasattr(self, 'run_dimrpad'):
            # Files for running DIMR
            self.write_dimrfiles()
        
        # Change mdu parameters
        for parameter, value in self.mdu_parameters.items():
            self.change_mdu_parameter(parameter, value)

        # Write laterals
        if any(self.dflowfmmodel.external_forcings.laterals):
            self.write_laterals()

    def _write_dict(self, f, dct, header,extra_linebreak=True):
        """
        Write a dictionary to file, without transforming with mapping
        """
        # Write header
        f.write(f'[{header}]\n')
        for key, value in dct.items():
            f.write(f'\t{key} = {value}\n')
        if extra_linebreak: f.write('\n')

    def writeMdFiles(self):
        """
        Copy MDU file form resouces.
        """
        srcMdu = os.path.join(os.path.dirname(__file__), 'resources', 'FMmdu.txt')
        targetMdu = os.path.join(self.output_dir, f'{self.name}.mdu')

        shutil.copy2(srcMdu, targetMdu)

        return True

    def write_roughness(self):

        # Write roughness
        roughnessfiles = []

        # Create a direcroty for the roughness files
        roughnessdir = 'roughness'
        if not os.path.exists(os.path.join(self.output_dir, roughnessdir)):
            os.mkdir(os.path.join(self.output_dir, roughnessdir))

        # Write each roughness definition to a file
        for i, (_, roughness) in enumerate(self.dflowfmmodel.network.roughness_definitions.items()):
            # Write roughness file for channels
            #file = os.path.join(roughnessdir, f'r{i+1:03d}.ini')
            file = f'{roughnessdir}/r{i+1:03d}.ini'
            roughnessfiles.append(file)
            with open(os.path.join(self.output_dir, file), 'w') as f:
                # header roughness
                self._write_header(f, filetype='roughness', fileversion=3.00)
                dct = {
                    'frictionId': roughness['name'],
                    'frictionType': roughness['code'],
                    'frictionValue': roughness['value'],
                }
                self._write_dict(f, dct=dct, header='Global')

        self.mdu_parameters['frictFile'] = ';'.join(roughnessfiles)

    def write_crosssection_locations(self):
        # Write cross section locations
        with open(os.path.join(self.output_dir, 'cross_section_locations.ini'), 'w') as f:
            self._write_header(f, 'crossLoc', fileversion=1.01)
            for _, series in self.dflowfmmodel.crosssections.crosssection_loc.items():
                # Write the cross section item
                self._write_dict(f, dct=series, header='CrossSection')

            # Write the default profile (if specified) to branches that do not have a cross section
            if self.dflowfmmodel.crosssections.default_definition is not None:
                # Find branches without profile
                no_css = self.dflowfmmodel.crosssections.get_branches_without_crosssection()
                branches = (self.dflowfmmodel.network.branches.loc[no_css, 'geometry'].length / 2).to_dict()
                for branchid, chainage in branches.items():
                    dct = {
                        'id': f'{branchid}_{chainage:.1f}',
                        'branchid': branchid,
                        'chainage': chainage,
                        'shift': self.dflowfmmodel.crosssections.default_definition_shift,
                        'definitionId': self.dflowfmmodel.crosssections.default_definition
                    }
                    # Write the cross section item
                    self._write_dict(f, dct=dct, header='CrossSection')

    def _write_header(self, f, filetype, fileversion, extra_linebreak=True):
        f.write(f"# Generated with D-HyDAMO in delftd3fmpy v.{self.version['number']} on {self.version['date']}.\n")
        f.write(f"# Tested and compatible with D-Flow FM {self.version['dfm_version']}, DIMRset {self.version['dimr_version']} and D-HYDRO suite 1D2D {self.version['suite_version']}.\n")
        f.write('# \n')
        f.write('[General]\n')
        f.write(f'fileVersion = {fileversion:.2f}\n')
        f.write(f'fileType = {filetype}\n')
        if extra_linebreak:
            f.write('\n')

    def write_crosssection_definitions(self):

        with open(os.path.join(self.output_dir, 'cross_section_definitions.ini'), 'w') as f:
            # header
            self._write_header(f, filetype='crossDef', fileversion=3.00)
            # definitions
            for _, dct in self.dflowfmmodel.crosssections.crosssection_def.items():
                self._write_dict(f, dct=dct, header='definition')

    def write_structures(self):

        filepath = os.path.join(self.output_dir, 'structure.ini')
        with open(filepath, 'w') as f:
            # write header
            self._write_header(f, filetype='structure', fileversion=2.00)

            # Culverts
            if any(self.dflowfmmodel.structures.culverts):
                for _, dct in self.dflowfmmodel.structures.culverts.items():
                    self._write_dict(f, dct=dct, header='Structure')

            # Weirs
            if any(self.dflowfmmodel.structures.weirs):
                for _, dct in self.dflowfmmodel.structures.weirs.items():
                    self._write_dict(f, dct=dct, header='Structure')

            # Bridges
            if any(self.dflowfmmodel.structures.bridges):
                for _, dct in self.dflowfmmodel.structures.bridges.items():
                    self._write_dict(f, dct=dct, header='Structure')

            # Pumps
            if any(self.dflowfmmodel.structures.pumps):
                for pump_id, dct in self.dflowfmmodel.structures.pumps.items():
                    self._write_dict(f, dct=dct, header='Structure')

            # Universal weirs
            if any(self.dflowfmmodel.structures.uweirs):
                for _, dct in self.dflowfmmodel.structures.uweirs.items():
                    self._write_dict(f, dct=dct, header='Structure')

            # Orifices
            if any(self.dflowfmmodel.structures.orifices):
                for _, dct in self.dflowfmmodel.structures.orifices.items():
                    self._write_dict(f, dct=dct, header='Structure')

            # Compound structures
            if any(self.dflowfmmodel.structures.compounds):
                for _, dct in self.dflowfmmodel.structures.compounds.items():
                    self._write_dict(f, dct=dct, header='structure')

        # Write structure time series
        for bcnd in self.dflowfmmodel.external_forcings.structures.itertuples():
            # Write time series for structure
            filename = f"{bcnd.type}_{bcnd.id}.tim"
            # Get time series as data
            data = list(zip(bcnd.time, bcnd.value))
            # Write file
            write_fm_file(file=os.path.join(self.output_dir, filename), data=data)


    def write_laterals(self):
        """
        Write laterals to files. Note that the .pol files are deprececated; locations are now in the ext-file.
        """

        if not 'refdate' in self.mdu_parameters.keys():
            raise KeyError('Specify reference date before writing laterals.')

        for name, dct in self.dflowfmmodel.external_forcings.laterals.items():

            if 'timeseries' in dct:
                # Determine starttime from mdu parameters
                starttime = datetime.datetime.strptime(str(self.mdu_parameters['refdate']), '%Y%m%d')
                starttime += datetime.timedelta(seconds=int(self.mdu_parameters['tstart']))

                # Write time series - this should probably be written to a .bc file
                series = dct['timeseries'].copy()
                series.index -= starttime
                data = np.c_[series.index.total_seconds().values / 60, series.values]

                refd = str(self.dflowfmmodel.mdu_parameters["refdate"])

                 # and write data to the bc file
                with open(os.path.join(self.output_dir, 'boundaries.bc'), 'a') as f:
                    f.write(f'\n[Forcing]\n'
                        f'name       = {name}\n'
                        f'function   = timeseries\n'
                        f'timeInterpolation = linear\n'
                        f'quantity   = time\n'
                        f'unit       = minutes since {refd[0:4]}-{refd[4:6]}-{refd[6:9]} 00:00:00\n'
                        f'quantity   = lateral_discharge\n'
                        f'unit       = m3/s\n'
                       )
                write_fm_file(file=os.path.join(self.output_dir, 'boundaries.bc'), data=data, mode='a')
                discharge_kw = f'boundaries.bc'
            else:
                discharge_kw = 'REALTIME'

            # Add to external forcing file
            with open(self.extfile_new, 'a') as f:
                dct = {'id':f'{name}',
                       'name':f'{name}',
                       'type':'discharge',
                       'locationType':'1d',
                       'branchId': f'{dct["branchid"]}',
                       'chainage': f'{dct["branch_offset"]}',
                       'discharge':f'{discharge_kw}'
                      }
                self._write_dict(f,dct,'Lateral')

    def write_boundary_conditions(self):
        """
        Write boundary conditions to (new format) ext file and the time series to a .bc file.
        """

        for bc in self.dflowfmmodel.external_forcings.boundaries.values():

            # Add boundary to ext file
            with open(self.extfile_new, 'a') as f:
                # check whether 2D rainfall has been applied to the 2D model
                if any('rainfall_2D' in key for key in self.dflowfmmodel.external_forcings.boundaries):

                    if bc['bctype'] == 'rainfallbnd':
                        # get file name of rainfall netcdf
                        rainfall_netcdf = self.dflowfmmodel.external_forcings.boundaries['rainfall_2D']['file_name']

                        #new: add precipitation netcdf for 2D mesh [Meteo] header
                        dct_met = {'quantity': 'rainfall',
                                   'forcingFile': os.path.basename(rainfall_netcdf),
                                   'forcingFileType': 'netcdf'}
                        self._write_dict(f, dct_met, 'Meteo')

                        shutil.copy2(rainfall_netcdf, self.output_dir)

                    if bc['bctype'] == 'waterlevelbnd' or bc['bctype'] == 'dischargebnd':

                        dct = {'quantity':f"{bc['bctype']}",'nodeId': bc['nodeid'],'forcingFile':'boundaries.bc'}
                        self._write_dict(f, dct, 'Boundary')

                        # Get time series
                        if bc['time'] is None:
                            data = float(bc['value'])
                            headerType='constant'
                        else:
                            data = list(zip(bc['time'], bc['value']))
                            headerType='timeseries'

                else:
                    dct = {'quantity':f"{bc['bctype']}",'nodeId': bc['nodeid'],'forcingFile':'boundaries.bc'}
                    self._write_dict(f, dct, 'Boundary')

                    # Get time series
                    if bc['time'] is None:
                        data = float(bc['value'])
                        headerType='constant'
                    else:
                        data = list(zip(bc['time'], bc['value']))
                        headerType='timeseries'

                refd = str(self.dflowfmmodel.mdu_parameters["refdate"])
    
                # and write data to the bc file
                with open(os.path.join(self.output_dir, 'boundaries.bc'), 'a') as f:
                    if headerType=='timeseries':
                        f.write(f'\n[Forcing]\n'
                            f'name       = {bc["nodeid"]}\n'
                            f'function   = timeseries\n'
                            f'timeInterpolation = linear\n'
                            f'quantity   = time\n'
                            f'unit       = minutes since {refd[0:4]}-{refd[4:6]}-{refd[6:9]} 00:00:00\n'
                            f'quantity   = {bc["bctype"]}\n'
                            f'unit       = {bc["unit"]}\n'
                           )
                    if headerType=='constant':
                        f.write(f'\n[Forcing]\n'
                            f'name       = {bc["nodeid"]}\n'
                            f'function   = constant\n'
                            f'quantity   = {bc["bctype"]}\n'
                            f'unit       = {bc["unit"]}\n'
                            f'{data}\n'
                           )
                if headerType=='timeseries':
                    write_fm_file(file=os.path.join(self.output_dir, 'boundaries.bc'), data=data, mode='a')

    def write_initial_conditions(self):

        """
        Two options exist: waterlevels per polygon or water depths per point. Both are written in the new format - an iniFieldFie is created pointing to the data.
        """

        # Create folder for initial conditions if it does not exist yet
        initcondfolder = 'initialconditions'

        # initial files are not in the ext-file in the new format...
        self.dflowfmmodel.mdu_parameters['IniFieldFile'] = initcondfolder+'/initialFields.ini'

        initcondpath = os.path.join(self.output_dir, initcondfolder)
        if not os.path.exists(initcondpath):
            os.mkdir(initcondpath)

        with open(os.path.join(initcondpath,'initialFields.ini'),'w') as f:
            self._write_header(f, 'iniField',2.00)

        # we have a water level definition
        if len(self.dflowfmmodel.external_forcings.initial_waterlevel_polygons)>0:
            if self.dflowfmmodel.external_forcings.initial_waterlevel_polygons['geometry'][0] is None:
                # no polygons defined - global definition
                with open(os.path.join(initcondpath,'initialWaterLevel.ini'),'w') as f:
                    self._write_header(f, '1dField',2.00)
                    dct = {'quantity': 'waterLevel','unit': 'm+NAP','value': self.dflowfmmodel.external_forcings.initial_waterlevel_polygons.waterlevel[0]}
                    self._write_dict(f, dct, 'Global')

                with open(os.path.join(initcondpath,'initialFields.ini'),'a') as f:
                     dct  = {'quantity': 'waterLevel',
                             'dataFileType': '1dField',
                             'dataFile': initcondfolder+'/initialWaterLevel.ini'
                             }
                     self._write_dict(f,dct,'Initial')
            else:
                for row in self.dflowfmmodel.external_forcings.initial_waterlevel_polygons.itertuples():
                    # Write pol file
                    write_fm_file(file=os.path.join(initcondpath, row.Index+".pol"),
                                      data=[np.vstack(row.geometry.exterior.coords[:])],
                                      names=[row.Index]
                                  )
                    with open(os.path.join(initcondpath,'initialFields.ini'),'a') as f:
                        dct  = {'quantity': 'waterLevel',
                                'dataFileType': 'polygon',
                                'dataFile': f'{row.Index}.pol',
                                'interpolationMethod': 'constant',
                                'value': f'{row.waterlevel}'
                                }
                        self._write_dict(f,dct,'Initial')

        # we have a water depth definition
        if len(self.dflowfmmodel.external_forcings.initial_waterdepth_polygons)>0:
            if self.dflowfmmodel.external_forcings.initial_waterdepth_polygons['geometry'][0] is None:
                # no polygons defined - global definition
                with open(os.path.join(initcondpath,'initialWaterDepth.ini'),'w') as f:
                    self._write_header(f, '1dField',2.00)
                    dct = {'quantity': 'waterDepth','unit': 'm','value': self.dflowfmmodel.external_forcings.initial_waterdepth_polygons.waterdepth[0]}
                    self._write_dict(f, dct, 'Global')

                with open(os.path.join(initcondpath,'initialFields.ini'),'a') as f:
                     dct  = {'quantity': 'waterDepth',
                             'unit': 'm',
                             'dataFileType': '1dField',
                             'dataFile': initcondfolder+'/initialWaterDepth.ini'
                             }
                     self._write_dict(f,dct,'Initial')
            else:
                for row in self.dflowfmmodel.external_forcings.initial_waterdepth_polygons.itertuples():
                    # Write pol file
                    write_fm_file(file=os.path.join(initcondpath, row.Index+".pol"),
                                      data=[np.vstack(row.geometry.exterior.coords[:])],
                                      names=[row.Index]
                                  )
                    with open(os.path.join(initcondpath,'initialFields.ini'),'a') as f:
                        dct  = {'quantity': 'waterDepth',
                                'unit': 'm',
                                'dataFileType': 'polygon',
                                'dataFile': f'{row.Index}.pol',
                                'interpolationMethod': 'constant',
                                'value': f'{row.waterdepth}'
                                }
                        self._write_dict(f,dct,'Initial')

    def objects_to_ldb(self, scalefactor=1.0):
        """
        Method to write weirs, bridges, pumps, universal weires, and cross sections to lbd
        """
        # Write cross sections
        coords = []
        names = []

        network = self.dflowfmmodel.network
        crosssections = self.dflowfmmodel.crosssections
        structures = self.dflowfmmodel.structures.as_dataframe(weirs=True, pumps=True, bridges=True, culverts=True, uweirs=True, orifices=True, compounds=True)

        for name, css in crosssections.crosssection_loc.items():
            sch_branch = network.schematised.at[css['branchid'], 'geometry']
            geo_branch = network.branches.at[css['branchid'], 'geometry']
            coords.append(orthogonal_line(
                line=sch_branch,
                offset=sch_branch.project(geo_branch.interpolate(css['chainage'])),
                width=10.0
            ))
            names.append(name)

        # # observation points
        # if len(self.dflowfmmodel.observation_points)>0:
        #     obs1d = self.dflowfmmodel.observation_points[self.dflowfmmodel.observation_points.locationType=='1d']

        #     for obspt in obs1d.itertuples():
        #         sch_branch = network.schematised.at[obspt.branch_id, 'geometry']
        #         geo_branch = network.branches.at[obspt.branch_id, 'geometry']
        #         offset = sch_branch.project(geo_branch.interpolate(obspt.branch_offset))
        #         names.append(obspt.code)
        #         coords.append(sch_branch.interpolate(offset).buffer(2.0).simplify(0.5).exterior.coords[:])

        if structures is not None:
            for structure in structures.itertuples():
                # compounds have no coordinates themselves
                if structure.type == 'compound':
                    continue
                sch_branch = network.schematised.at[structure.branchid, 'geometry']
                geo_branch = network.branches.at[structure.branchid, 'geometry']
                # Get offset on schematised branch
                offset = sch_branch.project(geo_branch.interpolate(structure.chainage))

                if structure.structype == 'pump':
                    # Get circle coordinates
                    coords.append(sch_branch.interpolate(offset).buffer(20.0).simplify(0.5).exterior.coords[:])
                    names.append(structure.Index)

                elif structure.structype == 'weir':
                    # Create orthogonal line
                    line = orthogonal_line(line=sch_branch, offset=offset, width=10.0)
                    # get coordinates for weir
                    ringcoords = LineString(line).parallel_offset(-2.0).coords[:] + LineString(line).parallel_offset(2.0).coords[::-1]
                    ringcoords += [ringcoords[0]]
                    coords.append(ringcoords)
                    names.append(structure.Index)

                elif structure.structype == 'bridge':
                    line1 = orthogonal_line(line=sch_branch, offset=offset-5., width=10.0)
                    line2 = orthogonal_line(line=sch_branch, offset=offset+5., width=10.0)
                    ringcoords = LineString(line1).coords[:] + LineString(line2).coords[::-1]
                    ringcoords += [ringcoords[0]]
                    coords.append(ringcoords)
                    names.append(structure.Index)

                elif structure.structype == 'uweir':
                    line1 = orthogonal_line(line=sch_branch, offset=offset-5., width=15.0)
                    line2 = orthogonal_line(line=sch_branch, offset=offset+5., width=1.0)
                    ringcoords = LineString(line1).coords[:] + LineString(line2).coords[::-1]
                    ringcoords += [ringcoords[0]]
                    coords.append(ringcoords)
                    names.append(structure.Index)

                elif structure.structype == 'orifice':
                    line1 = orthogonal_line(line=sch_branch, offset=offset+5., width=15.0)
                    line2 = orthogonal_line(line=sch_branch, offset=offset-5., width=1.0)
                    ringcoords = LineString(line1).coords[:] + LineString(line2).coords[::-1]
                    ringcoords += [ringcoords[0]]
                    coords.append(ringcoords)
                    names.append(structure.Index)

                elif structure.structype == 'culvert':
                    # Get circle coordinates
                    coords.append(sch_branch.interpolate(offset).buffer(5.0).simplify(0.5).exterior.coords[:])
                    names.append(structure.Index)

                elif structure.structype == 'orifice':
                    line1 = orthogonal_line(line=sch_branch, offset=offset, width=5.0)
                    line2 = orthogonal_line(line=LineString(line1), offset=offset, width=5.0)
                    coords.append(LineString(line1).coords[:]+LineString(line2).coords[:])
                    names.append(structure.Index)

                else:
                    raise TypeError('Structure type not recognized')

        # Write to file
        filename = 'structures.ldb'
        write_fm_file(os.path.join(self.output_dir, filename), data=coords, names=names)

        # Modify in mdu file
        self.mdu_parameters['landboundaryfile'] = filename

    def write_observation_points(self):
        """
        Methode to write observation points to file, and set
        the observation points file name in the mdu file.
        """
        if not self.dflowfmmodel.observation_points.empty:
            # Define file name
            filename = 'obspts_obs.ini'
            with open(os.path.join(self.output_dir,filename),'w') as f:
                self._write_header(f, filetype='obsPoints', fileversion=2.00)
                self.dflowfmmodel.observation_points.drop('geometry', axis=1, inplace=True)
                if self.dflowfmmodel.observation_points['locationType'].str.count("1d").sum()>0:
                    obs1d = self.dflowfmmodel.observation_points[self.dflowfmmodel.observation_points['locationType']=='1d']
                    if self.dflowfmmodel.observation_points['locationType'].str.count("2d").sum()>0:
                        obs1d.drop(['x','y'], axis=1, inplace=True)
                    obs1ddct = obs1d.to_dict(orient='row')
                    for dct in obs1ddct:
                        self._write_dict(f, dct=dct, header='ObservationPoint')
                if self.dflowfmmodel.observation_points['locationType'].str.count("2d").sum()>0:
                    obs2d = self.dflowfmmodel.observation_points[self.dflowfmmodel.observation_points['locationType']=='2d']
                    obs2d.drop(['branchId','chainage'], axis=1, inplace=True)
                    obs2ddct = obs2d.to_dict(orient='row')
                    for dct in obs2ddct:
                        self._write_dict(f, dct=dct, header='ObservationPoint')

            # Write to fm format file
            #obspt_data = [[row.geometry.x, row.geometry.y, row.name] for row in self.dflowfmmodel.observation_points.itertuples()]
            #write_fm_file(os.path.join(self.output_dir, filename), obspt_data)

            # Change mdu
            self.change_mdu_parameter('ObsFile', filename)
        else:
            self.change_mdu_parameter('ObsFile', '')

    def change_mdu_parameter(self, parameter, value):
        """
        Change parameter value in mdu file
        """
        with open(os.path.join(self.output_dir, f'{self.name}.mdu'), 'r') as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            if line.lower().startswith(parameter.lower()):
                items = re.split('=|#|$\s+', line)
                key, oldvalue = items[0], items[1]
                if not key.strip().lower() == parameter.lower():
                    continue
                lines[i] = line.replace('='+oldvalue, f'= {value}'.ljust(len(oldvalue)+1))
                break

        with open(os.path.join(self.output_dir, f'{self.name}.mdu'), 'w') as f:
            f.write(''.join(lines))
    
    def write_dimrfiles(self):        
        """
        Method to write the files required for running in DIMR
        """
        # run.bat
        with open(os.path.join(self.output_dir, 'run.bat'),'w') as f:
                 f.write('@ echo off\n')
                 f.write('set OMP_NUM_THREADS=2\n')
                 f.write('call \"'+self.run_dimrpad+'\"\n')
                 f.write('pause\n')

        # coupling XML
        FM_comp_name = 'DFM'        
        with open(os.path.join(self.output_dir, 'dimr_config.xml'),'w') as f:       
                 f.write('<?xml version="1.0" encoding="utf-8" standalone="yes"?>\n')
                 f.write('<dimrConfig xmlns="http://schemas.deltares.nl/dimr" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://schemas.deltares.nl/dimr http://content.oss.deltares.nl/schemas/dimr-1.2.xsd">\n')
                 f.write('\t<documentation>\n')
                 f.write('\t\t<fileVersion>1.2</fileVersion>\n')
                 f.write('\t\t<createdBy>D-HyDAMO in delft3dfmpy v.'+self.version['number']+'</createdBy>\n')
                 f.write("\t\t<creationDate>"+self.version['date']+'</creationDate>\n')
                 f.write('\t</documentation>\n\n')        
                 f.write('\t<control>\n')
                 f.write('\t\t\t\t<start name="'+FM_comp_name+'" />\n')
                 f.write('\t</control>\n\n')                                      
                 f.write('\t<component name="'+FM_comp_name+'">\n')
                 f.write('\t\t<library>dflowfm</library>\n')
                 f.write('\t\t<workingDir>.</workingDir>\n')
                 f.write('\t\t<inputFile>'+self.name+'.mdu</inputFile>\n')
                 f.write('\t</component>\n')                                  
                 f.write('</dimrConfig>\n')
        
    

def _format_row(row):
    # Create format string
    string = ''
    for element in row:
        if isinstance(element, str):
            string += element + ' '
        elif isinstance(element, int):
            string += '{:6d} '.format(element)
        elif isinstance(element, float):
            string += '{:12.6f} '.format(element)
    return string

def write_fm_file(file, data, names=None, mode='w'):
    string = ''
    if names is not None:
        for geometry, name in zip(data, names):
            string += '{}\n'.format(name)
            string += '{:6d} {:6d}\n'.format(*np.shape(geometry))
            string += '\n'.join(list(map(_format_row, geometry))) + '\n'

    else:
        string = '\n'.join(list(map(_format_row, data))) + '\n'

    with open(file, mode) as f:
        f.write(string)




import logging

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import LineString

from delft3dfmpy.core import checks, geometry
from delft3dfmpy.datamodels.common import ExtendedDataFrame, ExtendedGeoDataFrame

logger = logging.getLogger(__name__)

roughness_gml = {
    1: "Chezy",
    2: "Manning",
    3: "StricklerNikuradse",
    4: "Strickler",
    5: "WhiteColebrook",
    6: "deBosBijkerk"
}


def generate_pumps(pompen, sturing, gemalen):
    """
    Generate pumps from hydamo data. The function combines the pumps
    with its steering and pumping stations to pumps that can be imported
    by dflowfm.

    Note that HYDAMO provides to ways of controlling pumps.
    1. By linking the control to the pumping station, and the pumps to the pumping station
    2. By linking the control directly to the pumps itself.

    The pumping capacity is given in cubic meters per minute in Hydamo. Dflowfm requires
    cubic meters per second, so the capacity is divided by 60.
    
    Parameters
    ----------
    pompen : gpd.GeoDataFrame
        Geometry and attributes of pumps
    sturing : gpd.GeoDataFrame
        Attributes of steering
    gemalen : gpd.GeoDataFrame
        Geometry and attributes of pumping stations
    
    Returns
    -------
    pd.DataFrame
        DataFrame with pump attributes suitable for dflowfm
    """
    # Copy dataframe
    pumps_dfm = pompen.copy()

    # HyDAMO contains m3/min, while D-Hydro needs m3/s
    pumps_dfm['maximalecapaciteit'] /= 60

    # Add sturing to pumps
    for idx, pump in pumps_dfm.iterrows():

        # Find sturing for pump
        sturingidx = (sturing.codegerelateerdobject == idx).values

        # The pump and 'sturing' might be linked to the pumping station,
        # so first check if there are multiple pumps with one 'sturing'
        if not sturingidx.sum() == 1:
            gemaalidx = (gemalen.code == pump.codegerelateerdobject).values
            # If there als multiple pumping stations connected to one pump, raise an error
            if sum(gemaalidx) != 1:
                raise IndexError('Multiple pumping stations (gemalen) found for pump.')

            # Find the idx if the pumping station connected to the pump
            gemaalidx = gemalen.iloc[np.where(gemaalidx)[0][0]]['code']
            # Find the control for the pumping station (and thus for the pump)
            sturingidx = (sturing.codegerelateerdobject == gemaalidx).values

            assert sum(sturingidx) == 1

        # Get the control by index
        pump_control = sturing.iloc[np.where(sturingidx)[0][0]]

        if pump_control.doelvariabelecode != 1 and pump_control.doelvariabelecode != 'waterstand':
            raise NotImplementedError('Sturing not implemented for anything else than water level (1).')

        # Add levels for suction side
        pumps_dfm.at[idx, 'startlevelsuctionside'] = pump_control['bovenmarge'] 
        pumps_dfm.at[idx, 'stoplevelsuctionside'] = pump_control['ondermarge'] 
    
    return pumps_dfm

def generate_weirs(weirs, afsluitmiddel=None, sturing=None):
    """
    Generate weirs from Hydamo input

    Currently only simple weirs can be applied. From Hydamo the attributes 
    'laagstedoorstroomhoogte' and 'kruinbreedte' are used to define the weir dimensions.

    The function already contains code for handling more complex weirs,
    but this code is not reached for now.
    
    Parameters
    ----------
    weirs : gpd.GeoDataFrame
        GeoDataFrame with geometry and attributes for weirs.
    
    Returns
    -------
    pd.DataFrame
        DataFrame with weir attributes suitable for dflowfm
    """

    weirs_dfm = weirs.copy().astype('object')
    logger.info('Currently only simple weirs can be applied. From Hydamo the attributes \'laagstedoorstroomhoogte\' and \'kruinbreedte\' are used to define the weir dimensions.')

    return weirs_dfm

def generate_orifices(orifices, afsluitmiddel=None, sturing=None):
    """
    Generate orifices from Hydamo input

    Parameters
    ----------
    orifices : gpd.GeoDataFrame
        GeoDataFrame with geometry and attributes for weirs.
    
    Returns
    -------
    pd.DataFrame
        DataFrame with orifice attributes suitable for dflowfm
    """

    orifices_dfm = orifices.copy().astype('object')
    if 'maximaaldebiet' not in orifices_dfm:  
        orifices_dfm['uselimitflow'] = 'false'
        orifices_dfm['limitflow'] = 0.0
    else:
        orifices_dfm['uselimitflow'] = 'true'
        orifices_dfm['limitflow'] = orifices_dfm['maximaaldebiet']
    return orifices_dfm


def generate_uweirs(uweirs, yz_profiles=None, parametrised_profiles=None):
   """
   Generate universal weirs from Hydamo input

   Crest level is determined from the Hydamo laagstedoorstroomhoogte attribute. The (relative) profile is determined from the crossection with codegerelateerdobject pointing to the universal weir.
    
   Parameters
   ---------
   uweirs : gpd.GeoDataFrame
       GeoDataFrame with geometry and attributes for universal weirs.
    
   Returns
   -------
   pd.DataFrame
       DataFrame with universal weir attributes suitable for dflowfm

   """
   
   uweirs_dfm = uweirs.copy().astype('object')
   uweirs_dfm['crosssection'] = [{} for _ in range(len(uweirs_dfm))]

   for uweir in uweirs.itertuples():    
       
       # first search in yz-profiles
       prof=np.empty(0)
       if yz_profiles is not None:
            if 'codegerelateerdobject' in yz_profiles:  
                prof = yz_profiles[yz_profiles['codegerelateerdobject']==uweir.code]   
                if not prof.empty:                    
                    counts = len(prof.geometry[0].coords[:])
                    xyz = np.vstack(prof.geometry[0].coords[:])
                    length = np.r_[0, np.cumsum(np.hypot(np.diff(xyz[:, 0]), np.diff(xyz[:, 1])))]
                    yzvalues = np.c_[length, xyz[:, -1]-np.min(xyz[:,-1])]            
                
       if (prof.empty) & (parametrised_profiles is not None):
            if 'codegerelateerdobject' in parametrised_profiles:  
                # if not found, check the parametrised profiles
                prof = parametrised_profiles[parametrised_profiles['codegerelateerdobject']==uweir.code]           
                nulls = pd.isnull(prof[['bodembreedte', 'bodemhoogtebenedenstrooms', 'bodemhoogtebovenstrooms','taludhellinglinkerzijde','taludhellingrechterzijde','hoogteinsteeklinkerzijde','hoogteinsteekrechterzijde']]).any(axis=1).values
                if nulls:
                   raise ValueError(f'Insufficient fields avaialable in parametrised profile defintion {prof.code}.')
                bodemhoogte = float((prof.bodemhoogtebenedenstrooms + prof.bodemhoogtebovenstrooms)/2.)
                zvalues = [ float(prof.hoogteinsteeklinkerzijde-bodemhoogte), 0., 0., float( prof.hoogteinsteekrechterzijde-bodemhoogte )]
                yvalues = list( np.cumsum( [0.,float((prof.hoogteinsteeklinkerzijde-bodemhoogte)/prof.taludhellinglinkerzijde),float(prof.bodembreedte), float((prof.hoogteinsteekrechterzijde-bodemhoogte)/prof.taludhellingrechterzijde) ] ) )           
                yzvalues = list(zip(yvalues,zvalues))
                counts = len(zvalues)
           
       if len(prof)==0:
           # return an error it is still not found
           raise ValueError(f'{uweir.code} is not found in any cross-section.')
    
       uweirs_dfm.at[uweir.Index, 'numlevels'] = counts
       uweirs_dfm.at[uweir.Index, 'yvalues'] = ' '.join([f'{yz[0]:7.3f}' for yz in yzvalues])
       uweirs_dfm.at[uweir.Index, 'zvalues'] = ' '.join([f'{yz[1]:7.3f}' for yz in yzvalues])        
             
   return uweirs_dfm
            
        # # Check levels
        # if weir.laagstedoorstroomhoogte >= weir.hoogstedoorstroomhoogte:
        #     weirs.at[weir.Index, 'weirtype'] = 'weir'

        # else:
        #     weirs.at[weir.Index, 'weirtype'] = 'weir'
        #     # The universal weir is not supported yet in D-Hydro!
        #     # self.weirs.at[weir.Index, 'weirtype'] = 'universal weir'

        #     # Create y,z-values
        #     yzvalues = [
        #         (-0.5 * weir.kruinbreedte, weir.hoogstedoorstroomhoogte),
        #         (-0.5 * weir.hoogstedoorstroombreedte, weir.hoogstedoorstroomhoogte),
        #         (-0.5 * weir.laagstedoorstroombreedte, weir.laagstedoorstroomhoogte),
        #         (0.5 * weir.laagstedoorstroombreedte, weir.laagstedoorstroomhoogte),
        #         (0.5 * weir.hoogstedoorstroombreedte, weir.hoogstedoorstroomhoogte),
        #         (0.5 * weir.kruinbreedte, weir.hoogstedoorstroomhoogte)
        #     ]

        #     # Remove duplicate values
        #     counts = [yzvalues.count(yz) for yz in yzvalues]
        #     while any([c > 1 for c in counts]):
        #         yzvalues.remove(yzvalues[counts.index(max(counts))])
        #         counts = [yzvalues.count(yz) for yz in yzvalues]

        #     # Add values
        #     weirs.at[weir.Index, 'ycoordinates'] = ' '.join([f'{yz[0]:7.3f}' for yz in yzvalues])
        #     weirs.at[weir.Index, 'zcoordinates'] = ' '.join([f'{yz[1]:7.3f}' for yz in yzvalues])
        #     weirs.at[weir.Index, 'levelscount'] = len(yzvalues)

def generate_bridges(bridges, yz_profiles=None, parametrised_profiles=None):
    """
    Generate bridges from Hydamo input

    Parameters
    
    ----------
    bridges : gpd.GeoDataFrame
        GeoDataFrame with geometry and attributes for bridges.
    
    Returns
    -------
    pd.DataFrame
        DataFrame with weir attributes suitable for dflowfm
    """

    bridges_dfm = bridges.copy().astype('object')
    bridges_dfm['crosssection'] = [{} for _ in range(len(bridges_dfm))]
    bridges_dfm['bedlevel'] = [{} for _ in range(len(bridges_dfm))]
    
    for bridge in bridges.itertuples():        
        # first search in yz-profiles
        prof = yz_profiles[yz_profiles['codegerelateerdobject']==bridge.code]
        if len(prof) > 0:
            bedlevel = np.min([c[2] for c in prof.geometry[0].coords[:]])  
            profile_id=prof.code.values[0]
        else:
            # if not found, check the parametrised profiles
            prof = parametrised_profiles[parametrised_profiles['codegerelateerdobject']==bridge.code]
            bedlevel = (prof['bodemhoogtebovenstrooms'] + prof['bodemhoogtebenedenstrooms'])/2.
           
        if len(prof)==0:
            # return an error it is still not found
            raise ValueError(f'{bridge.code} is not found in any cross-section.')
        
        profile_id=prof.code.values[0]
        bridges_dfm.at[bridge.Index, 'crosssection'] = profile_id
        bridges_dfm.at[bridge.Index, 'bedlevel'] = float(bedlevel)
             
    return bridges_dfm
          
def generate_culverts(culverts,afsluitmiddel):

    culverts_dfm = culverts.copy()
    culverts_dfm['crosssection'] = [{} for _ in range(len(culverts_dfm))]
        
    for culvert in culverts.itertuples():

        # Generate cross section definition name
        if culvert.vormcode == 1 or culvert.vormcode == 'rond' or culvert.vormcode == 5 or culvert.vormcode == 'ellipsvormig':
            crosssection = {'shape': 'circle', 'diameter': culvert.hoogteopening}
            
        elif culvert.vormcode == 3 or culvert.vormcode == 'rechthoekig' or culvert.vormcode == 99 or culvert.vormcode == 'onbekend':
            crosssection = {'shape': 'rectangle', 'height': culvert.hoogteopening, 'width': culvert.breedteopening, 'closed': 1}
        
        else:
            crosssection = {'shape': 'circle', 'diameter': 0.40}
            print(f'Culvert {culvert.code} has an unknown shape: {culvert.vormcode}. Applying a default profile (round - 40cm)')
        
        # Set cross section definition
        culverts_dfm.at[culvert.Index, 'allowedflowdir'] = 'both'
        culverts_dfm.at[culvert.Index, 'valveonoff'] = 0
        culverts_dfm.at[culvert.Index, 'numlosscoeff'] = 0
        culverts_dfm.at[culvert.Index, 'valveopeningheight'] = 0
        culverts_dfm.at[culvert.Index, 'relopening'] = 0
        culverts_dfm.at[culvert.Index, 'losscoeff'] = 0
        # check whether an afsluitmiddel is present and take action dependent on its settings
        if afsluitmiddel is not None:
            if not afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code].empty:
                if len(afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code])!=1:
                    raise IndexError(f'Multiple (or no) instances of afsluitmiddel associated with culvert {culvert.code}')
                if int(afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code]['soortafsluitmiddelcode'])==5:
                    culverts_dfm.at[culvert.Index, 'allowedflowdir'] = 'positive'
                if int(afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code]['soortafsluitmiddelcode'])==4:
                    culverts_dfm.at[culvert.Index, 'valveonoff'] = 1
                    culverts_dfm.at[culvert.Index, 'valveopeningheight'] = float(afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code]['hoogte'])
                    culverts_dfm.at[culvert.Index, 'numlosscoeff'] = 1
                    culverts_dfm.at[culvert.Index, 'relopening'] = float(afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code]['hoogte'])/culvert.hoogteopening
                    culverts_dfm.at[culvert.Index, 'losscoeff'] = float(afsluitmiddel[afsluitmiddel.codegerelateerdobject==culvert.code]['afvoercoefficient'])
        culverts_dfm.at[culvert.Index, 'crosssection'] = crosssection

    return culverts_dfm
    
def move_structure(struc, struc_dict, branch, offset):
    """
    Function the move a structure if needed for a compound event.
    
    Parameters
    ----------
    struc : string
        current sub-structure id
    struc_dict : dict
        dict with all structures of a certain type
    branch : string
        branch id of the first structure in the compound
    offset : float
        chainage of the first structure in the compound

    Returns
    -------
    Dict with shifted coordinates.

    """
    branch2 = struc_dict[struc]['branchid']   
    if branch2!=branch:
       logger.warning(f'Structures of not on the same branche. Moving structure {struc} to branch {branch}.')
    struc_dict[struc]['branchid'] = branch
    struc_dict[struc]['chainage'] = offset
    return struc_dict

def generate_compounds(idlist, structurelist, structures):
    # probably the coordinates should all be set to those of the first structure (still to do)        
    compounds_dfm = ExtendedDataFrame(required_columns=['code','structurelist'])
    compounds_dfm.set_data(pd.DataFrame(np.zeros((len(idlist),3)), columns=['code','numstructures','structurelist'], dtype='str'),index_col='code')
    compounds_dfm.index = idlist    
    for ii,compound in enumerate(compounds_dfm.itertuples()):
        compounds_dfm.at[compound.Index, 'code'] = idlist[ii]
        compounds_dfm.at[compound.Index, 'numstructures'] = len(structurelist[ii])
        
        # check the substructure coordinates. If they do not coincide, move subsequent structures to the coordinates of the first
        for s_i, struc in enumerate(structurelist[ii]):            
            if s_i == 0:                
                # find out what type the first structure it is and get its coordinates
                if struc in structures.pumps.keys():
                    branch = structures.pumps[struc]['branchid']
                    offset = structures.pumps[struc]['chainage']
                elif struc in structures.weirs.keys():
                    branch = structures.weirs[struc]['branchid']
                    offset = structures.weirs[struc]['chainage']
                elif struc in structures.uweirs.keys():
                    branch = structures.uweirs[struc]['branchid']
                    offset = structures.uweirs[struc]['chainage']
                elif struc in structures.culverts.keys():
                    branch = structures.culverts[struc]['branchid']
                    offset = structures.culverts[struc]['chainage']                    
                elif struc in structures.bridges.keys():
                    branch = structures.bridges[struc]['branchid']
                    offset = structures.bridges[struc]['chainage']                    
                elif struc in structures.orifices.keys():
                    branch = structures.orifices[struc]['branchid']
                    offset = structures.orifices[struc]['chainage']                                        
                else:
                    raise IndexError('Structure id not found. Make sure all other structures have been added to the model.')
            else:
                # move a subsequent structure to the location of the first
                 if struc in structures.pumps.keys():                    
                     structures.pumps = move_structure(struc, structures.pumps, branch, offset)
                 if struc in structures.weirs.keys():                    
                     structures.weirs = move_structure(struc, structures.weirs, branch, offset)
                 if struc in structures.uweirs.keys():                    
                     structures.uweirs = move_structure(struc, structures.uweirs, branch, offset)
                 if struc in structures.culverts.keys():                    
                     structures.culverts = move_structure(struc, structures.culverts, branch, offset)
                 if struc in structures.bridges.keys():                    
                     structures.bridges = move_structure(struc, structures.bridges, branch, offset)
                 if struc in structures.orifices.keys():                    
                     structures.orifices = move_structure(struc, structures.orifices, branch, offset)                 
                                     
        compounds_dfm.at[compound.Index, 'structurelist'] = ';'.join([f'{s}' for s in structurelist[ii]])
    
    return compounds_dfm

def dwarsprofiel_to_yzprofiles(crosssections, branches):
    """
    Function to convert hydamo cross sections 'dwarsprofiel' to
    dflowfm input.
    d
    Parameters
    ----------
    crosssections : gpd.GeoDataFrame
        GeoDataFrame with x,y,z-coordinates of cross sections
    
    Returns
    -------
    dictionary
        Dictionary with attributes of cross sections, usable for dflowfm
    """
    cssdct = {}

    for css in crosssections.itertuples():
        # The cross sections from hydamo are all yz profiles
        
       # if css.Index == 'prof_RS1-DP-27341':
       #     print('stop')
        # Determine yz_values
        xyz = np.vstack(css.geometry.coords[:])
        length = np.r_[0, np.cumsum(np.hypot(np.diff(xyz[:, 0]), np.diff(xyz[:, 1])))]
        yz = np.c_[length, xyz[:, -1]]
        # the GUI cannot cope with identical y-coordinates. Add 1 cm to a 2nd duplicate.
        yz[:,0] = np.round(yz[:,0],3)
        for i in range(1,yz.shape[0]):
            if yz[i,0]==yz[i-1,0]:
                yz[i,0] +=0.01
                
        # determine thalweg
        if branches is not None:
            branche_geom = branches[branches.code==css.branch_id].geometry.values        
            thalweg_xyz = css.geometry.intersection(branche_geom[0]).coords[:][0]                
            # and the Y-coordinate of the thalweg
            thalweg = np.hypot( thalweg_xyz[0]-xyz[0,0], thalweg_xyz[1]-xyz[0,1])
        else: 
            thalweg = 0.0
        
        # Add to dictionary
        cssdct[css.code] = {
            'branchid': css.branch_id,
            'chainage': css.branch_offset,
            'yz': yz,
            'thalweg':thalweg,
            'ruwheidstypecode': css.ruwheidstypecode,
            'ruwheidswaarde': css.ruwheidswaarde
        }
    
    return cssdct

def parametrised_to_profiles(parametrised, branches):
    """
    Generate parametrised cross sections for all branches,
    or the branches missing a cross section.

    Parameters
    ----------
    parametrised : pd.DataFrame
        GeoDataFrame with geometries and attributes of parametrised profiles.
    branches : list
        List of branches for which the parametrised profiles are derived

    Returns
    -------
    dictionary
        Dictionary with attributes of cross sections, usable for dflowfm
    """

    checks.check_argument(parametrised, 'parametrised', (pd.DataFrame, gpd.GeoDataFrame))
    checks.check_argument(branches, 'branches', (list, tuple))

    # Find
    if len(branches) != 0:
         parambranches = ExtendedGeoDataFrame(geotype=LineString, columns = parametrised.columns.tolist()+['css_type'])
         parambranches.set_data(parametrised[np.isin(parametrised.code, branches)], index_col='code', check_columns=True)
    else:
        parambranches = parametrised
     #   
    #else:
        #parambranches = parametrised.reindex(columns=parametrised.requiredcolumns + ['css_type'])
    #    parambranches = parametrised[np.isin(parametrised.code,branches)]# ['css_type'])
   
   
    # Drop profiles for which not enough data is available to write (as rectangle)
    #nulls = pd.isnull(parambranches[['bodembreedte', 'bodemhoogtebenedenstrooms', 'bodemhoogtebovenstrooms']]).any(axis=1).values
    #parambranches = parambranches.drop(ExtendedGeoDataFrame(geotype=LineString), parambranches.index[nulls], index_col='code',axis=0)
    #parambranches.drop(parambranches.index[nulls], inplace=True)

    # Determine characteristics
    botlev = (parambranches['bodemhoogtebenedenstrooms'] + parambranches['bodemhoogtebovenstrooms']) / 2.0
    dh1 = parambranches['hoogteinsteeklinkerzijde'] - botlev
    dh2 = parambranches['hoogteinsteekrechterzijde'] - botlev
    parambranches['height'] = (dh1 + dh2) / 2.0
    parambranches['bottomlevel'] = botlev

    # Determine maximum flow width and slope (both needed for output)
    parambranches['maxflowwidth'] = parambranches['bodembreedte'] + parambranches['taludhellinglinkerzijde'] * dh1 + parambranches['taludhellingrechterzijde'] * dh2
    parambranches['slope'] = (parambranches['taludhellinglinkerzijde'] + parambranches['taludhellingrechterzijde']) / 2.0

    # Determine profile type
    parambranches.loc[:, 'css_type'] = 'trapezium'
    nulls = pd.isnull(parambranches[parametrised.required_columns]).any(axis=1).values
    parambranches.loc[nulls, 'css_type'] = 'rectangle'

    cssdct = {}
    for branch in parambranches.itertuples():
        # Determine name for cross section
        if branch.css_type == 'trapezium':
            cssdct[branch.Index] = {
                'type': branch.css_type,
                'slope': round(branch.slope, 1),
                'maximumflowwidth': round(branch.maxflowwidth, 1),
                'bottomwidth': round(branch.bodembreedte, 3),
                'closed': 0,
                'thalweg': 0.0,
                'ruwheidstypecode': int(branch.ruwheidstypecode) if isinstance(branch.ruwheidstypecode, float) else branch.ruwheidstypecode,
                'ruwheidswaarde': branch.ruwheidswaarde,
                'bottomlevel': branch.bottomlevel
            }
        elif branch.css_type == 'rectangle':
            cssdct[branch.Index] = {
                'type': branch.css_type,
                'height': 5.0,
                'width': round(branch.bodembreedte, 3),
                'closed': 0,
                'thalweg': 0.0,
                'ruwheidstypecode': int(branch.ruwheidstypecode) if isinstance(branch.ruwheidstypecode, float) else branch.ruwheidstypecode,
                'ruwheidswaarde': branch.ruwheidswaarde,
                'bottomlevel': branch.bottomlevel
            }

    return cssdct

def generate_boundary_conditions(boundary_conditions, schematised):
    """
    Generate boundary conditions from hydamo 'randvoorwaarden' file.

    Parameters
    ----------
    boundary_conditions: gpd.GeoDataFrame
        geodataframe with the locations and properties of the boundary conditions
    schematised : gpd.GeoDataFrame
        geodataframe with the schematised branches
    
    Returns
    -------
    dictionary
        Dictionary with attributes of boundary conditions, usable for dflowfm
    """
    bcdct = {}

    for bndcnd in boundary_conditions.itertuples():

        # Find nearest branch for geometry
        # extended_line = geometry.extend_linestring(
        #     line=schematised.at[bndcnd.branch_id, 'geometry'], near_pt=bndcnd.geometry, length=1.0)

        # # Create intersection line for boundary condition
        # bcline = LineString(geometry.orthogonal_line(line=extended_line, offset=0.1, width=0.1))

        if bndcnd.typerandvoorwaardecode in [0, 'waterstand']:
            bctype = 'waterlevel'
        elif bndcnd.typerandvoorwaardecode in [1, 'debiet']:
            bctype = 'discharge'

        # Add boundary condition
        bcdct[bndcnd.code] = {
            'code': bndcnd.code,
            'bctype': bctype,#+'bnd',
            'value': bndcnd.waterstand if not np.isnan(bndcnd.waterstand) else bndcnd.debiet,
            'time': None,
            'geometry': bndcnd.geometry,#bcline,
            'filetype': 9,
            'operand': 'O',
            'method': 3,
            'branchid': bndcnd.branch_id
        }

    return bcdct



import os
import pickle

import geopandas as gpd
import pandas as pd

from delft3dfmpy.datamodels.common import ExtendedDataFrame, ExtendedGeoDataFrame
from shapely.geometry import LineString, Point, Polygon
import logging

class OSM:
    """
    OpenStreetMap model
    """

    def __init__(self, extent_file=None, data_columns=None, proj_crs = None, logger=logging):

        # Read geometry to clip data
        self.logger = logger
        self.crs_out = proj_crs

        if extent_file is not None:
            self.logger.debug(f'extent file found, reading from {extent_file}')
            self.clipgdf = gpd.read_file(extent_file)
            self.clipgeo = self.clipgdf.unary_union

            if self.crs_out is not None:
                if self.crs_out!=self.clipgdf.crs:
                    self.clipgdf.to_crs(self.crs_out, inplace=True)
        else:
            self.logger.debug(f'No extent file found, assuming the entire file is needed.')
            self.clipgdf = None
            self.clipgeo = None



        # Get required columns of OSM data
        self.data_columns = data_columns

        # Create standard dataframe for network, cross sections, orifices, weirs
        # FIXME: check available columns and required columns for the OSM data, and apply these here
        self.branches = ExtendedGeoDataFrame(geotype=LineString, required_columns=self.get_columns('branches'))

        # FIXME: ensure that all required parameterised properties are provided. I can imagine this is a matter of making
        # several parameterised profiles for different profile types (e.g. trapezoidal, rectangular, circular, etc.)
        self.profiles = ExtendedGeoDataFrame(geotype=LineString, required_columns=self.get_columns('crosssections'))

        # FIXME: ensure that all culvert types and properties can be handled. We probably have circular and box-shaped culverts, sometimes with multiple openings
        self.culverts = ExtendedGeoDataFrame(geotype=LineString, required_columns=self.get_columns('culverts'))

        # # FIXME: not sure what laterals in this context mean, but I don't think we need it at this stage.
        # self.laterals = ExtendedGeoDataFrame(geotype=Point, required_columns=[
        #     'code',
        #     'geometry'
        # ])
        #

    def get_columns(self, key):
        cols = [x.strip() for x in self.data_columns[key].replace(' ', '').split(',')]
        return cols

    def to_pickle(self, filename, overwrite=False):
        # Check if path exists
        if os.path.exists(filename) and not overwrite:
            raise FileExistsError(f'File "{filename}" already exists.')

        # Dump object
        with open(filename, 'wb') as handle:
            pickle.dump(self, handle)

    @classmethod
    def from_pickle(cls, filename):
        # Read object
        with open(filename, 'rb') as handle:
            loaded_cls = pickle.load(handle)

        return loaded_cls

